# This file is part of pyDC software
# Copyright 2003-2004 Anakim Border <aborder@users.sourceforge.net>
#
# pyDC is released under the terms of GPL licence.

from JobViewer import *
from JobViewerInfo import *
from os import sep
from Util import *
import wx
import wx.gizmos

from DCUserList import UserDir, UserFile
from DCQueue import ItemAlreadyAddedError
from DCQueueItem import *
from DCItem import DCItem
from DCLocalList import DCLocalList
import DCWorker

class FileListViewer(JobViewer):
	def __init__(self, parent, id, info):
		JobViewer.__init__(self, parent, id, info)
		self.continueSearch = False
		self.menuId = -1

		# Controls & layout
		searchBox = wx.StaticBox(self, -1, 'Search')
		self.searchText = wx.TextCtrl(self, 1, style = wx.TE_PROCESS_ENTER)
		self.searchButton = wx.Button(self, 2, 'Start')
		
		self.tree = wx.gizmos.TreeListCtrl(self, 3, style = wx.TR_HAS_BUTTONS | wx.TR_HIDE_ROOT)
		close = wx.Button(self, 4, 'Close')
		
		self.tree.AddColumn('File')
		self.tree.AddColumn('Size')
		self.tree.SetMainColumn(0)
		self.tree.SetColumnAlignment(1, wx.LIST_FORMAT_RIGHT)

		# Layout
		hLayout = wx.StaticBoxSizer(searchBox, wx.HORIZONTAL)
		hLayout.Add(self.searchText, 1, wx.EXPAND | wx.RIGHT, 5)
		hLayout.Add(self.searchButton, 0, wx.EXPAND | wx.RIGHT, 2)
		
		vLayout = wx.BoxSizer(wx.VERTICAL)
		vLayout.Add(hLayout, 0, wx.EXPAND | wx.LEFT | wx.RIGHT, 2)
		vLayout.Add(self.tree, 1, wx.EXPAND | wx.TOP, 5)
		vLayout.Add(close, 0, wx.ALIGN_CENTER | wx.BOTTOM | wx.TOP, 5)
		
		self.tree.SetColumnWidth(1, self.tree.GetCharWidth()*len('999.99 GB'))
		self.SetSizer(vLayout)
		self.Layout()

		# Events
		wx.EVT_TEXT(self, 1, self.onPatternUpdate)
		wx.EVT_TEXT_ENTER(self, 1, self.onSearch)
		wx.EVT_BUTTON(self, 2, self.onSearch)
		wx.EVT_SIZE(self.tree, self.onTreeResize)
		wx.EVT_TREE_ITEM_RIGHT_CLICK(self, 3, self.onResultOptions)
		wx.EVT_MENU_RANGE(self.tree, 0, 501, self.onPopupSelection)
		wx.EVT_TREE_ITEM_ACTIVATED(self.tree, 3, self.onQueueItem)
		wx.EVT_BUTTON(self, 4, self.onClose)

		# Tree population
		j = self.info.getJob()
		j.sort()
		id = self.tree.AddRoot('root')
		for item in j.getList():
			self.populateBranch(id, item)

	def onPatternUpdate(self, event):
		self.continueSearch = False
	
	def onSearch(self, event):
		pattern = self.searchText.GetValue()
		if len(pattern) == 0: return
		
		id = self.tree.GetSelection()
		if not id.IsOk(): self.continueSearch = False
		if not self.continueSearch: id = self.tree.GetRootItem()
		
		while True:
			id = self.nextItem(id)
			if id == None: break
		
			item = self.tree.GetPyData(id)
			if item == None:
				match = False
			else:
				match = True
				for p in pattern.lower().split(' '):
					name = item.name.lower()
					if name.find(p.strip()) == -1:
						match = False
						break
			
			if match:
				self.tree.EnsureVisible(id)
				self.tree.SelectItem(id)
				self.continueSearch = True
				return
		
		self.continueSearch = False
		showErrorMessage(self, 'No more results matching your query', 'Search error')

	def onTreeResize(self, event):
		w = event.GetSize().GetWidth() - self.tree.GetColumnWidth(1)
		if wx.Platform != '__WXMSW__':
			w -= wx.SystemSettings_GetSystemMetric(wx.SYS_VSCROLL_X)
		self.tree.SetColumnWidth(0, w)
		event.Skip()
	
	def onResultOptions(self, event):
		pnt = event.GetPoint()
		obj = self.tree.GetItemData(event.GetItem()).GetData()
		self.tree.SelectItem(event.GetItem())
		
		menu = wx.Menu()
		menu.Append(500, 'Queue')
		
		if isinstance(obj, UserFile):
			pos, altMenu, alternates = 0, wx.Menu(), []
			items = DCWorker.getWorker().getQueue().getItems()
			for item in items:
				if item.getStatus() != DCQueueItem.STATUS_COMPLETED and obj.size == item.getSize():
					altMenu.Append(pos, basename(item.getPath()))
					alternates.append(item)
					pos += 1
			if pos: menu.AppendMenu(1, 'Source for', altMenu)

		self.tree.PopupMenu(menu, pnt)

		try:
			if self.menuId == -1:
				return
			
			elif self.menuId == 500:
				busy = wx.BusyInfo('Queueing items...')
				wx.Yield()
				self.queueItem(obj)
			
			else:
				try:
					alternates[self.menuId].addSource(obj)
				except ExistingSourceError:
					pass
				except ItemNotMatching:
					showErrorMessage(self, 'Source is not matching.')
			
		except ItemAlreadyAddedError:
			showErrorMessage(self, 'Item already queued.')
		
		self.menuId = -1
		
	def onPopupSelection(self, event):
		self.menuId = event.GetId()

	def onQueueItem(self, event):
		data = self.tree.GetItemData(event.GetItem()).GetData()
		
		if isinstance(data, UserFile):
			self.queueItem(data)
		else:
			busy = wx.BusyInfo('Queueing items...')
			wx.Yield()
			self.queueDir(data)
	
	def onClose(self, event):
		self.viewerDestroy(False)
	
	def getTitle(self):
		if isinstance(self.info.getJob(), DCLocalList):
			return DCWorker.getWorker().getSettings().getNick()
		else:
			return self.info.getJob().getUser().nick

	def update(self):
		pass

	#=================#
	# Private section #
	#=================#
	
	def nextItem(self, id):
		if self.tree.ItemHasChildren(id):
			cookie = 0
			cid, cookie = self.tree.GetFirstChild(id, cookie)
			return cid
		
		sid = self.tree.GetNextSibling(id)
		if sid.IsOk(): return sid
		
		while True:
			id = self.tree.GetItemParent(id)
			if not id.IsOk(): break
			
			nid = self.tree.GetNextSibling(id)
			if nid.IsOk(): return nid
		
		return None

	def populateBranch(self, parent, item):
		data = wx.TreeItemData(item)
		if isinstance(item, UserDir):
			id = self.tree.AppendItem(parent, item.name, data = data)
			for i in item.children:
				self.populateBranch(id, i)
		else:
			id = self.tree.AppendItem(parent, item.name, data = data)
			self.tree.SetItemText(id, formatSize(item.size), 1)
			
	def queueItem(self, item, prefix=None):
		i = DCItem()
		i.userNick, i.path, i.size = item.fullInfo()
		
		if prefix:
			try:    localPath = i.path[i.path.rindex('\\')+1:]
			except: localPath = i.path
			localPath = prefix + sep + localPath
		else:
			localPath = None
		
		try:
			DCWorker.getWorker().getQueue().addItem(i, localPath)
		except ItemAlreadyAddedError:
			showErrorMessage(self, 'Item already queued.')
	
	def queueDir(self, dir, prefix=None):
		getMainWnd().enableQueueRefresh(False)
	
		if prefix:
			prefix = prefix + sep + dir.name
		else:
			prefix = dir.name
		
		for child in dir.getChildren():
			if isinstance(child, UserFile):
				self.queueItem(child, prefix)
			else:
				self.queueDir(child, prefix)
		
		getMainWnd().enableQueueRefresh(True)

class FileListViewerInfo(JobViewerInfo): pass
