# This file is part of pyDC software
# Copyright 2002-2003 Anakim Border <aborder@users.sourceforge.net>
#
# pyDC is released under the terms of GPL licence.

from ViewerListener import *
from Util import getMainWnd, getListWidth, guiEvent, caseFilter, formatSize
import wx

from DCHub import *
from DCHubEventListener import *
import DCWorker

class HubsPanel(wx.Panel, ViewerListener, DCHubEventListener):
	def __init__(self, parent, id):
		wx.Panel.__init__(self, parent, id)
		self.hubs = []
		
		self.sortAlgos = [ (caseFilter, False), (None, False), (None, True), (None, True) ]
		self.sortColumn = 2
		self.sortReverse = False
		self.timestamp = 0

		# Controls
		self.list = wx.ListCtrl(self, 1, style = wx.LC_REPORT | wx.LC_SINGLE_SEL | wx.LC_VRULES)
		self.list.InsertColumn(0, 'Name')
		self.list.InsertColumn(1, 'Address')
		self.list.InsertColumn(2, 'Users', wx.LIST_FORMAT_RIGHT)
		self.list.InsertColumn(3, 'Share', wx.LIST_FORMAT_RIGHT)
		
		# Layout
		vLayout = wx.BoxSizer(wx.VERTICAL)
		vLayout.Add(self.list, 1, wx.EXPAND | wx.TOP, 5)
		
		self.list.Freeze()
		self.list.InsertStringItem(0, '###')
		self.list.SetStringItem(0, 2, '99999')
		self.list.SetStringItem(0, 3, '999.99 GB')
		self.list.SetColumnWidth(2, wx.LIST_AUTOSIZE)
		self.list.SetColumnWidth(3, wx.LIST_AUTOSIZE)
		self.list.DeleteItem(0)
		self.list.Thaw()

		self.SetSizer(vLayout)
		self.Layout()
		
		# Events
		wx.EVT_LIST_COL_CLICK(self, 1, self.setSortAlgo)
		wx.EVT_SIZE(self.list, self.onListResize)
		wx.EVT_LIST_ITEM_SELECTED(self.list, 1, self.onHubActivation)
		self.Connect(1, 1, wx.wxEVT_NULL, self.addHub)
		self.Connect(2, 2, wx.wxEVT_NULL, self.removeHub)
		self.Connect(3, 3, wx.wxEVT_NULL, self.chatMsg)
		self.Connect(4, 4, wx.wxEVT_NULL, self.askPassword)

	def setSortAlgo(self, event):
		col = event.GetColumn()
		if self.sortColumn == col:
			self.sortReverse = not self.sortReverse
		else:
			self.sortColumn = col
		self.update()
	
	def onListResize(self, event):
		colWidths = [ self.list.GetColumnWidth(col) for col in range(0,4) ]
		width = getListWidth(self.list) - colWidths[2] - colWidths[3]
		if width > 0:
			scaleFactor = float(colWidths[0]) / (colWidths[0] + colWidths[1])
			w = int(width * scaleFactor)
			self.list.SetColumnWidth(0, w)
			self.list.SetColumnWidth(1, width - w)
		event.Skip()

	def onHubActivation(self, event):
		try: getMainWnd().viewJob(self.hubs[event.GetIndex()])
		except IndexError: pass

	def onNewHub(self, hub):
		hub.registerListener(self)
		hub.registerListener(self, 'Channel')
		self.AddPendingEvent(guiEvent(1, hub))
	
	def onChannelClosed(self, hub):
		if not hub.isAlive():
			self.AddPendingEvent(guiEvent(2, hub))
	
	def onLogUpdate(self, hub, row):
		if row[0] == DCHub.LOG_PRIVCHAT:
			self.AddPendingEvent(guiEvent(3, (hub, row[2])))
	
	def onUserPassword(self, hub):
		self.AddPendingEvent(guiEvent(4, hub))
	
	def onViewerClosed(self, viewer, stoppedJob):
		if stoppedJob:
			self.AddPendingEvent(guiEvent(2, viewer.getInfo().getJob()))
	
	def addHub(self, event):
		self.hubs.append(event.data)
		self.update()

	def removeHub(self, event):
		try:
			pos = self.hubs.index(event.data)
			event.data.deregisterListener(self, 'Channel')
			event.data.deregisterListener(self)
			del self.hubs[pos]
			self.list.DeleteItem(pos)
		except ValueError:
			pass
	
	def askPassword(self, event):
		hub = event.data
		dialog = wx.TextEntryDialog(self, 'Your nick is registered. Please enter the password:', 'Password for %s' % hub.getName())
		if dialog.ShowModal() == wx.ID_OK:
			hub.setPassword(dialog.GetValue())
		else:
			hub.stop()
	
	def chatMsg(self, event):
		hub, row = event.data

		# TODO: consistency check
		idx = row.index('>')
		nick = row[1:idx]
		msg = row[idx+2:]

		chat = getMainWnd().getChat(nick, hub)
		if chat: chat.appendMsg(msg)
	
	def update(self):
		snap = self.sort()
		
		self.list.Freeze()
		count = self.list.GetItemCount()
		for pos, s in enumerate(snap):
			self.hubs[pos] = s[4]
			
			if pos < count:
				self.list.SetStringItem(pos, 0, s[0])
			else:
				self.list.InsertStringItem(pos, s[0])
			
			self.list.SetStringItem(pos, 1, '%s:%d' % s[1])
			self.list.SetStringItem(pos, 2, str(s[2]))
			self.list.SetStringItem(pos, 3, formatSize(s[3]))

		self.list.Thaw()

	#=================#
	# Private section #
	#=================#
	
	def sort(self):
		snap = [ (h.getName(), h.getAddress(), h.getUserNum(), h.getSize(), h) for h in self.hubs ]
		filter, reverse = self.sortAlgos[self.sortColumn]
		if self.sortReverse: reverse = not reverse
		
		if filter != None:
			aux = [ (filter(s[self.sortColumn]), s) for s in snap ]
		else:
			aux = [ (s[self.sortColumn], s) for s in snap ]
			
		aux.sort()
		if reverse: aux.reverse()
		return [ a[1] for a in aux ]
