# This file is part of pyDC software
# Copyright 2002-2004 Anakim Border <aborder@users.sourceforge.net>
#
# pyDC is released under the terms of GPL licence.

from EventGenerator import *
from JobViewer import *
from JobViewerInfo import *
from Util import *
import os.path
import wx

from DCSearch import *
from DCSearchResult import *
from DCSearchEventListener import *
from DCUserList import DCUserList
from DCQueue import ItemAlreadyAddedError
from DCQueueItem import *
import DCWorker

class SearchViewer(JobViewer, DCSearchEventListener):
	RESULT_ACCESS = (DCSearchResult.getUserNick, DCSearchResult.getPath, DCSearchResult.getType, DCSearchResult.getSize, DCSearchResult.getFreeSlots, DCSearchResult.getHubName)
	
	def __init__(self, parent, id, info):
		JobViewer.__init__(self, parent, id, info)
		
		self.sortAlgos = [ (caseFilter, False), (self.pathFilter, False), (caseFilter, False), (None, True), (None, True), (caseFilter, False) ]
		self.sortColumn = 1
		self.sortReverse = False
		
		self.results = self.info.getJob().getResults()
		self.newResults = []
		self.timer = wx.Timer(self, 7)
		self.menuId = -1

		# Controls
		patternLabel = wx.StaticText(self, -1, 'Pattern:')
		self.patternText = wx.TextCtrl(self, 1, style = wx.TE_PROCESS_ENTER)
		sizeLabel = wx.StaticText(self, -1, 'Size:')
		self.limitType = wx.ComboBox(self, 2, style = wx.CB_DROPDOWN | wx.CB_READONLY)
		self.limitSize = wx.TextCtrl(self, 3, style = wx.TE_PROCESS_ENTER | wx.TE_RIGHT)
		sizeUnit = wx.StaticText(self, -1, 'KB')
		self.start = wx.Button(self, 4, 'Start')
		self.list = wx.ListCtrl(self, 5, style = wx.LC_REPORT | wx.LC_VRULES | wx.LC_SINGLE_SEL)
		close = wx.Button(self, 6, 'Close')
		
		self.limitType.Append('no limit')
		self.limitType.Append('at least')
		self.limitType.Append('at most')
		self.limitType.Select(0)

		self.list.InsertColumn(0, 'User')
		self.list.InsertColumn(1, 'File')
		self.list.InsertColumn(2, 'Type')
		self.list.InsertColumn(3, 'Size', wx.LIST_FORMAT_RIGHT)
		self.list.InsertColumn(4, 'Slots')
		self.list.InsertColumn(5, 'Hub')
		
		# Layout
		sLayout = wx.BoxSizer(wx.HORIZONTAL)
		sLayout.Add(self.limitType, 0, wx.RIGHT, 5)
		sLayout.Add(self.limitSize, 1, wx.EXPAND | wx.RIGHT, 5)
		sLayout.Add(sizeUnit, 0, wx.ALIGN_CENTER_VERTICAL)
		
		gs = wx.FlexGridSizer(cols = 2, hgap = 5, vgap = 5)
		gs.AddGrowableCol(1)
		gs.Add(patternLabel, 0, wx.ALIGN_CENTER_VERTICAL)
		gs.Add(self.patternText, 1, wx.EXPAND)
		gs.Add(sizeLabel, 0, wx.ALIGN_CENTER_VERTICAL)
		gs.Add(sLayout, 1, wx.EXPAND)
		
		vLayout = wx.BoxSizer(wx.VERTICAL)
		vLayout.Add(gs, 0, wx.EXPAND | wx.LEFT | wx.TOP | wx.RIGHT, 5)
		vLayout.Add(self.start, 0, wx.ALIGN_CENTER | wx.TOP, 5)
		vLayout.Add(self.list, 1, wx.EXPAND | wx.LEFT | wx.TOP | wx.RIGHT, 5)
		vLayout.Add(close, 0, wx.ALIGN_CENTER | wx.BOTTOM | wx.TOP, 5)
		
		self.list.Freeze()
		self.list.InsertStringItem(0, '###')
		self.list.SetStringItem(0, 2, 'www')
		self.list.SetStringItem(0, 3, '999.99 GB')
		self.list.SetStringItem(0, 4, '99/99')
		self.list.SetColumnWidth(2, wx.LIST_AUTOSIZE)
		self.list.SetColumnWidth(3, wx.LIST_AUTOSIZE)
		self.list.SetColumnWidth(4, wx.LIST_AUTOSIZE)
		self.list.DeleteItem(0)
		self.list.Thaw()

		self.SetSizer(vLayout)
		self.Layout()

		# Events
		wx.EVT_COMBOBOX(self, 2, self.onType)
		wx.EVT_BUTTON(self, 4, self.onStart)
		wx.EVT_BUTTON(self, 6, self.onClose)
		wx.EVT_LIST_COL_CLICK(self, 5, self.setSortAlgo)
		wx.EVT_SIZE(self.list, self.onListResize)
		wx.EVT_LIST_ITEM_RIGHT_CLICK(self, 5, self.onResultOptions)
		wx.EVT_LIST_ITEM_ACTIVATED(self, 5, self.onResultSelected)
		wx.EVT_MENU_RANGE(self.list, 0, 502, self.onPopupSelection)
		wx.EVT_TIMER(self, 7, self.onTimer)
		self.Connect(1, 1, wx.wxEVT_NULL, self.addResult)
		
		self.limitSize.Enable(False)
		self.patternText.SetFocus()
		wx.EVT_TEXT_ENTER(self, 1, self.onStart)
		wx.EVT_TEXT_ENTER(self, 3, self.onStart)
	
	def getTitle(self):
		pattern = self.info.getJob().getPattern()
		if pattern != None:
			return pattern
		else:
			return '[New search]'
	
	def onType(self, event):
		enabled = self.limitType.GetSelection() != 0
		if not enabled: self.limitSize.SetValue('')
		self.limitSize.Enable(enabled)

	def onStart(self, event):
		self.info.pattern = self.patternText.GetValue()
		if len(self.info.pattern) == 0:
			return showErrorMessage(self, 'No pattern present.', 'Search error')
		
		self.info.limitType = self.limitType.GetSelection()
		if self.info.limitType != 0:
			try:
				self.info.limitSize = int(self.limitSize.GetValue())
			except ValueError:
				return showErrorMessage(self, 'Invalid size limit value.', 'Search error')
		
		self.patternText.Enable(False)
		self.limitType.Enable(False)
		self.limitSize.Enable(False)
		self.start.Enable(False)

		job = self.info.getJob()
		job.setPattern(self.info.pattern)
		if self.info.limitType > 0:
			job.setSizeLimit([DCSearch.LIMIT_NONE, DCSearch.LIMIT_ATLEAST, DCSearch.LIMIT_ATMOST][self.info.limitType], self.info.limitSize*1024)
		job.start()
		
		self.lock.acquire()
		for listener in self.getListeners():
			listener.onViewerTitle(self, self.info.pattern)
		self.lock.release()

	def onClose(self, event):
		self.info.getJob().stop()
		self.viewerDestroy(True)

	def setSortAlgo(self, event):
		col = event.GetColumn()
		if self.sortColumn == col:
			self.sortReverse = not self.sortReverse
		else:
			self.sortColumn = col
		self.sort()
		
		self.list.Freeze()
		self.list.DeleteAllItems()
		for pos, result in enumerate(self.results):
			self.insertResult(pos, result)
		self.list.Thaw()
		
	def onListResize(self, event):
		colWidths = [ self.list.GetColumnWidth(col) for col in range(0,6) ]
		width = getListWidth(self.list) - sum(colWidths[2:5])
		if width > 0:
			s = colWidths[0] + colWidths[1] + colWidths[5]
			w1 = int(width * (float(colWidths[0]) / s))
			w2 = int(width * (float(colWidths[1]) / s))
			self.list.SetColumnWidth(0, w1)
			self.list.SetColumnWidth(1, w2)
			self.list.SetColumnWidth(5, width - w1 - w2)
		event.Skip()

	def onResultSelected(self, event):
		try:
			DCWorker.getWorker().getQueue().addItem(self.results[event.GetIndex()])
		except ItemAlreadyAddedError:
			showErrorMessage(self, 'Item already queued.')

	def onResultOptions(self, event):
		pnt = event.GetPoint()
		result = self.results[event.GetIndex()]
		self.list.SetItemState(event.GetIndex(), wx.LIST_STATE_SELECTED, wx.LIST_STATE_SELECTED)

		menu = wx.Menu()
		menu.Append(500, 'Queue')
		menu.Append(501, 'Full path')
		menu.Append(502, 'Get file list')
		
		pos, altMenu, alternates = 0, wx.Menu(), []
		for item in DCWorker.getWorker().getQueue().getItems():
			if item.getStatus() != DCQueueItem.STATUS_COMPLETED and result.size == item.getSize():
				altMenu.Append(pos, os.path.basename(item.getPath()))
				alternates.append(item)
				pos += 1
		if pos: menu.AppendMenu(1, 'Source for', altMenu)

		self.list.PopupMenu(menu, pnt)

		try:
			if self.menuId == -1:
				return
			
			elif self.menuId == 500:
				DCWorker.getWorker().getQueue().addItem(result)
			
			elif self.menuId == 501:
				wx.MessageBox(result.path)
			
			elif self.menuId == 502:
				try:
					w = DCWorker.getWorker()
					h = w.getHubByName(result.getHubName())
					if h == None: raise ValueError
					
					u = h.getUserByNick(result.getUserNick())
					if u == None: raise ValueError
					
					l = DCUserList(u)
				except ValueError:
					showErrorMessage(self, 'No such user.')
			
			else:
				try: alternates[self.menuId].addSource(result)
				except ExistingSourceError: pass
				except ItemNotMatching:
					showErrorMessage(self, 'Source is not matching.')
		
		except ItemAlreadyAddedError:
			showErrorMessage(self, 'Item already queued.')
		
		self.menuId = -1

	def onPopupSelection(self, event):
		self.menuId = event.GetId()

	def onNewResult(self, search, result):
		self.AddPendingEvent(guiEvent(1, result))
		
	def addResult(self, event):
		self.newResults.append(event.data)
		if not self.timer.IsRunning():
			self.timer.Start(1000, True)
	
	def onTimer(self, event=None):
		self.timer.Stop()
		
		res = self.newResults
		self.newResults = []
		if len(res) == 0: return
		
		self.results += res
		self.sort()
		
		posList = [ (self.results.index(r), r) for r in res ]
		posList.sort()
		self.list.Freeze()
		for pos, result in posList:
			self.insertResult(pos, result)
		self.list.Thaw()

	def update(self):
		pass

	#=================#
	# Private section #
	#=================#
	
	def pathFilter(self, path):
		return basename(path)
	
	def sort(self):
		filter, reverse = self.sortAlgos[self.sortColumn]
		if self.sortReverse: reverse = not reverse
		
		if filter != None:
			aux = [ (filter(self.RESULT_ACCESS[self.sortColumn](r)), r) for r in self.results ]
		else:
			aux = [ (self.RESULT_ACCESS[self.sortColumn](r), r) for r in self.results ]
		
		aux.sort()
		if reverse: aux.reverse()
		self.results = [ a[1] for a in aux ]

	def insertResult(self, pos, result):
		self.list.InsertStringItem(pos, result.userNick)
		self.list.SetStringItem(pos, 1, basename(result.path))
		self.list.SetStringItem(pos, 2, result.type.lower())
		self.list.SetStringItem(pos, 3, formatSize(result.size))
		self.list.SetStringItem(pos, 4, '%d/%d' % (result.freeSlots, result.slots))
		self.list.SetStringItem(pos, 5, result.hubName)

class SearchViewerInfo(JobViewerInfo): pass
