# This file is part of pyDC software
# Copyright 2002-2004 Anakim Border <aborder@users.sourceforge.net>
#
# pyDC is released under the terms of GPL licence.

import os
import errno
from Job import Job
from socket import *
from threading import Lock
from time import time
from DNS import *

class AsyncSocket(object, Job):
	STATUS_READY = 0
	STATUS_RESOLVING = 1
	STATUS_CONNECTING = 2
	STATUS_CONNECTED = 3
	STATUS_CLOSING = 4
	STATUS_CLOSED = 5
	
	CONNECTION_TIMEOUT = 30
	DISCONNECTION_TIMEOUT = 5
	DEFAULT_WINDOW = 0xffff
	
	def __init__(self, sock = None):
		self._status = self.STATUS_READY
		self.obj = None
		
		self.lock = Lock()
		self.timer = 0
		self.window = self.DEFAULT_WINDOW
		
		self.inbuf = ''
		self.outbuf = ''
		self.outlen = 0
		self.incount = 0
		self.outcount = 0
		
		if sock:
			if sock.getpeername() == None:
				raise error, (error.EBADF, 'Unconnected socket')
			
			sock.setblocking(0)
			self.obj = sock
			self.setStatus(self.STATUS_CONNECTED)

	def connect(self, host):
		if self._status != self.STATUS_READY:
			raise error, (errno.EBADF, 'Socket not ready')

		self.host = host
		try:
			inet_aton(host[0])
			self.setStatus(self.STATUS_CONNECTING)
		except:
			self.obj = DNSResolver(host[0])
			self.setStatus(self.STATUS_RESOLVING)

	def close(self):
		self.setStatus(self.STATUS_CLOSING)
		
	def poll(self):
		raise error, (errno.ENOTCONN, 'Socket not connected')

	def recv(self, size = -1):
		self.lock.acquire()
		if size == -1:
			data = self.inbuf
			self.inbuf = ''
		else:
			data = self.inbuf[:size]
			self.inbuf = self.inbuf[size:]
		self.lock.release()
		
		return data
	
	def unrecv(self, data):
		self.lock.acquire()
		self.inbuf = data + self.inbuf
		self.lock.release()

	def send(self, data):
		self.lock.acquire()
		self.outlen += len(data)
		self.outbuf += data
		self.lock.release()

	def getpeername(self):
		try:
			return self.obj.getpeername()
		except:
			return None
	
	def getsockname(self):
		try:
			return self.obj.getsockname()
		except:
			return None

	def isAlive(self):
		self.lock.acquire()
		res = self._status > self.STATUS_READY and self._status <= self.STATUS_CLOSING
		self.lock.release()
		return res

	def isConnected(self):
		return self.poll == self.handleConnected

	def getStatus(self):
		return self._status
	
	def getWindow(self):
		return self._window

	def setWindow(self, size):
		if size < 0:
			self._window = self.DEFAULT_WINDOW
		else:
			self._window = size

	def getBytesToSend(self):
		return self.outlen

	def getBytesSent(self):
		return self.outcount

	def getBytesReceived(self):
		return self.incount

	#=================#
	# Private section #
	#=================#
	
	def handleResolve(self):
		try:
			if not self.obj.poll():
				return False
		except DNSError, e:
			self.setStatus(self.STATUS_CLOSING)
			raise error, (errno.EADDRNOTAVAIL, 'DNS error: %s' % e[0])
		
		self.host = (self.obj.resolved, self.host[1])
		self.timer = time()
		return self.setStatus(self.STATUS_CONNECTING)
	
	def handleConnecting(self):
		if self.timer == 0:
			self.obj = socket(AF_INET, SOCK_STREAM)
			self.obj.setblocking(False)
			self.timer = time()
		elif time() - self.timer > self.CONNECTION_TIMEOUT:
			self.setStatus(self.STATUS_CLOSING)
			raise error, (errno.ETIMEDOUT, 'Timeout while connecting')
		
		try:
			self.obj.connect(self.host)
		except error, e:
			failure = True
			if self.isWouldBlock(e[0]):
				return False
			elif os.name == 'nt':
				if e[0] == errno.WSAEISCONN:
					failure = False
				elif e[0] == errno.WSAEINVAL:
					return False
			
			if failure:
				del self.host
				self.setStatus(self.STATUS_CLOSING)
				raise e
		
		del self.host
		return self.setStatus(self.STATUS_CONNECTED)
	
	def handleConnected(self):
		try:
			self.lock.acquire()
				
			# Receive
			try:
				data = self.obj.recv(self.window)
				dataLen = len(data)
				if dataLen == 0:
					return self.setStatus(self.STATUS_CLOSING, False)
				
				self.inbuf += data
				self.incount = dataLen
			except error, e:
				if self.isWouldBlock(e[0]):
					self.incount = 0
				else:
					self.setStatus(self.STATUS_CLOSING, False)
					raise e
	
			# Send
			if self.outlen > 0:
				try:
					sent = self.obj.send(self.outbuf)
	
					self.outbuf = self.outbuf[sent:]
					self.outlen -= sent
					self.outcount = sent
				except error, e:
					if self.isWouldBlock(e[0]):
						self.outcount = 0
					else:
						self.setStatus(self.STATUS_CLOSING, False)
						raise e
			else:
				self.outcount = 0
		
		finally:
			self.lock.release()
		
		return False
	
	def handleClosing(self):
		if isinstance(self.obj, DNSResolver):
			try:
				if not self.obj.poll():
					return False
			except DNSError:
				pass
			
			self.obj = None
			return self.setStatus(self.STATUS_CLOSED)
		
		else:
			if self.timer == 0:
				try:
					self.obj.shutdown(1)
					self.timer = time()
					close = False
				except error:
					close = True
			else:
				close = time() - self.timer > self.DISCONNECTION_TIMEOUT
			
			if not close:
				try:
					if len(self.obj.recv(self.DEFAULT_WINDOW)) != 0:
						return False
				except error: pass
			
			self.obj.close()
			self.obj = None
			
			return self.setStatus(self.STATUS_CLOSED)
	
	def setStatus(self, status, lock = True):
		if lock: self.lock.acquire()
		if status > self._status:
			self._status = status
			self.timer = 0
			self.poll = [ None, self.handleResolve, self.handleConnecting, self.handleConnected, self.handleClosing, lambda:True ][status]
		if lock: self.lock.release()
		
		return self._status != self.STATUS_CLOSED

	def isWouldBlock(self, code):
		if os.name == "posix":
			if code in (errno.EAGAIN, errno.EALREADY, errno.EINPROGRESS, errno.EINTR):
				return True
		elif os.name == "nt":
			if code == errno.WSAEWOULDBLOCK:
				return True

		return False

	status = property(getStatus)
	peername = property(getpeername)
	alive = property(isAlive)
	connected = property(isConnected)
	window = property(getWindow, setWindow)
	bytesToSend = property(getBytesToSend)
