# This file is part of pyDC software
# Copyright 2003-2004 Anakim Border <aborder@users.sourceforge.net>
#
# pyDC is released under the terms of GPL licence.

from AsyncSocket import *
from EventGenerator import *
from Job import *
from threading import Lock
from weakref import proxy
import socket

class Channel(object, Job, EventGenerator):
	STATUS_READY = 0
	STATUS_CONNECTING = 1
	STATUS_ESTABLISHED = 2
	STATUS_CLOSING = 3

	def __init__(self):
		EventGenerator.__init__(self)
		self.addInterface('Channel')

		self.status = self.STATUS_READY
		self.connected = False
		self.opLock = Lock()

		self._addr = None
		self.delay = None
		self.sock = None
		self.handler = None
		self.cleanup = []

	def attach(self, sock):
		if not sock.isConnected():
			raise InvalidSocket
	
		try:
			self.opLock.acquire()
			if self.status != self.STATUS_READY or self.delay:
				raise InvalidStatus
	
			self.status = self.STATUS_ESTABLISHED
			self.connected = True
			self._addr = sock.getpeername()
			self.sock = sock
		finally:
			self.opLock.release()
		
		self.onConnection()

	def connect(self, addr = None, defaultPort = 0, delay = 0):
		#TODO: add state check?
		try:
			self.opLock.acquire()
			
			if addr:
				self._addr = self.parseAddress(addr, defaultPort)
			elif self._addr == None:
				raise InvalidAddress
			
			self.delay = time() + delay
			
			if self.status != self.STATUS_READY:
				self.status = self.STATUS_CLOSING
		finally:
			self.opLock.release()

	def stop(self):
		self.opLock.acquire()
		if self.status != self.STATUS_READY or self.delay != None:
			self.status = self.STATUS_CLOSING
		self.delay = None
		self.opLock.release()

	def poll(self):
		try:
			if self.status == self.STATUS_ESTABLISHED:
				try:
					if self.handler:
						self.handler.prepoll()
				except Exception:
					self.setStatus(self.STATUS_CLOSING)
					return False

				if not self.sock.isConnected():
					self.setStatus(self.STATUS_CLOSING)
				self.sock.poll()

				try:
					if self.handler:
						self.handler.poll()
				except Exception:
					self.setStatus(self.STATUS_CLOSING)
					return False

				return False

			elif self.status == self.STATUS_READY:
 				if self.delay == None:
					return True
				elif time() < self.delay:
					return False
				
				self.opLock.acquire()
				self.sock = AsyncSocket()
				self.sock.connect(self._addr)
				
				self.status = self.STATUS_CONNECTING
				self.delay = None
				self.connected = False
				self.opLock.release()
				
				return False

			elif self.status == self.STATUS_CONNECTING:
				self.sock.poll()
				if self.sock.isConnected():
					self.setStatus(self.STATUS_ESTABLISHED)
					self.connected = True
					self.onConnection()

				return False

			elif self.status == self.STATUS_CLOSING:
				while len(self.cleanup):
					cl = self.cleanup.pop()
					cl[0](cl[1])
				
				if self.sock:
					if self.sock.isConnected():
						self.sock.close()
					
					if not self.sock.poll():
						return False

				self.opLock.acquire()
				self.status = self.STATUS_READY
				self.sock = None
				self.opLock.release()

				if self.connected:
					self.onDisconnection()
				
				if self.delay == None:
					self.lock.acquire()
					for listener in self.getListeners('Channel'):
						listener.onChannelClosed(self)
					self.lock.release()

				return False

			else:
				return True

		except socket.error, e:
			self.setStatus(self.STATUS_CLOSING)
			self.onSocketError(e)

	def isAlive(self):
		return self.status in (self.STATUS_CONNECTING, self.STATUS_ESTABLISHED, self.STATUS_CLOSING) or \
		       (self.status == self.STATUS_READY and self.delay != None)

	def setHandler(self, handler):
		try:
			self.opLock.acquire()
			if self.handler:
				self.handler.onDetach()
	
			self.handler = handler
		finally:
			self.opLock.release()
		
		try:
			self.handler.onAttach(self)
		except Exception:
			self.setStatus(self.STATUS_CLOSING)
	
	def setWindow(self, size):
		self.sock.setWindow(size)
	
	def registerCleanup(self, func, data):
		self.cleanup.append((func, data))
	
	def clearCleanup(self):
		self.cleanup = []

	#=================#
	# Private members #
	#=================#

	def setStatus(self, status):
		self.opLock.acquire()
		if status > self.status:
			self.status = status
		self.opLock.release()

	def parseAddress(self, addr, defaultPort):
		if isinstance(addr, str) or isinstance(addr, unicode):
			pos = addr.find(':')
			if pos != -1:
				try:
					port = int(addr[pos+1:])
					if port <= 0: raise InvalidAddress
					return (addr[:pos], port)
				except ValueError:
					raise InvalidAddress
			else:
				if defaultPort <= 0: raise InvalidAddress
				return (addr, defaultPort)
		else:
			if len(addr) != 2: raise InvalidAddress
			if not (isinstance(addr[0], str) or isinstance(addr[0], unicode)) or not isinstance(addr[1], int):
				raise InvalidAddress
			return addr
			
	# Internal events
	def onConnection(self):
		pass

	def onDisconnection(self):
		pass

	def onSocketError(self, e):
		pass

class ChannelHandler:
 	def __init__(self):
 		self.channel = None

	def prepoll(self):
		pass

	def poll(self):
		pass

	def onAttach(self, channel):
		self.channel = proxy(channel)

	def onDetach(self):
		pass

class ChannelEventListener:
	def onChannelClosed(self, channel):
		pass

class InvalidStatus(Exception): pass
class InvalidAddress(Exception): pass
class InvalidSocket(Exception): pass
