# This file is part of pyDC software
# Copyright 2002-2004 Anakim Border <aborder@users.sourceforge.net>
#
# pyDC is released under the terms of GPL licence.

from os import access, R_OK, W_OK, makedirs, unlink
from os.path import dirname, exists, getsize, isfile
from time import time
from zlib import decompressobj

from Channel import ChannelHandler
from DCChannel import DCChannelHandler
import DCWorker

class DCDownloadSetupHandler(DCChannelHandler):
	HANDSHAKE_TIMEOUT = 30

	def __init__(self):
		cmds = { 'Error' :      ( 0, DCDownloadSetupHandler.onError),
		         'MaxedOut' :   ( 0, DCDownloadSetupHandler.onMaxedOut),
	             'FileLength' : ( 1, DCDownloadSetupHandler.onFileLength),
		         'Sending' :    ( 0, DCDownloadSetupHandler.onSending),
		         'Failed' :     (-1, DCDownloadSetupHandler.onFailed) }
		DCChannelHandler.__init__(self, cmds)
		self.block = None

	def onAttach(self, channel):
		DCChannelHandler.onAttach(self, channel)

		queue = DCWorker.getWorker().getQueue()
		item, source = queue.getItemByNick(self.channel.getRemoteNick())
		if item == None:
			raise ValueError
		
		if queue.MAX_CONCURRENT_BLOCKS > 1 and queue.AVOID_OLD_CLIENTS and not self.channel.supportsSegments():
			item.removeSource(self.channel.getRemoteNick(), False)
			raise ValueError, '[download] avoiding old client'

		block = item.getBlock(self.channel)
		if block == None:
			raise ValueError

		self.rollback = self.channel.DOWNLOAD_ROLLBACK
		self.timestamp = time()
		
		size = 0
		if exists(block.path):
			if not isfile(block.path) or not access(block.path, R_OK | W_OK):
				raise ValueError #TODO: let the user know about this

			size = getsize(block.path)
			if size == block.size:
				item.setStatusCompleted(self, block)
				return self.channel.nextItem()
			elif size > block.size:
				unlink(block.path)

		self.channel._item = item
		self.block = block

		oldSize = size
		if self.rollback > 0:
			self.rollback = min(size, self.rollback)
			size -= self.rollback

		restart = block.offset + size
		self.channel._sizeRemaining = block.size - size
		
		if self.channel.supportsSegments() and block.size != -1:
			self.channel._compressed = True
			if self.channel.supports('GetZBlock'):
				cmd = 'GetZBlock'
			else:
				cmd = 'GetTestZBlock'
			
			self.channel.sendCmd(cmd, str(restart), str(self.channel._sizeRemaining), source.getPath(self.channel))
		else:
			self.channel._compressed = False
			self.channel.sendCmd('Get', source.getPath(self.channel) + '$' + str(restart+1))

	def poll(self):
		if time() - self.timestamp > self.HANDSHAKE_TIMEOUT:
			raise ValueError, '[download] handshake timeout'

		DCChannelHandler.poll(self)

	def onInvalidCommand(self, cmd):
		item = self.channel.getItem()
		item.removeSource(self.channel.getRemoteNick(), False, item.CAUSE_XFER_ERROR)
		raise ValueError, '[download] invalid command'

	def onUnknownCommand(self, cmd):
		raise ValueError

	def onChatMessage(self, cmd):
		raise ValueError

	def onError(self):
		item = self.channel.getItem()
		item.removeSource(self.channel.getRemoteNick(), False, item.CAUSE_XFER_ERROR)
		self.channel.nextItem()

	def onMaxedOut(self):
		raise ValueError, '[download] maxed out'

	def onFileLength(self, length):
		size = long(length)
		item = self.channel.getItem()

		if item.getSize() == -1:
			# We don't know yet how big this item is; it's a file list
			item.setSize(size)
			self.channel._sizeRemaining = size

		elif item.getSize() != size:
			item.removeSource(self.channel.getRemoteNick(), False, item.CAUSE_XFER_ERROR)
			return self.channel.nextItem()
		
		if self.startDownload(item):
			self.channel.sendCmd('Send')
	
	def onSending(self):
		self.startDownload(self.channel._item)
	
	def onFailed(self, cmd):
		item = self.channel.getItem()
		item.removeSource(self.channel.getRemoteNick(), False, item.CAUSE_XFER_ERROR)
		return self.channel.nextItem()

	def startDownload(self, item):
		# Try to acquire the running state on the block
		if not item.setStatusRunning(self.channel, self.block):
			self.channel.nextItem()
			return False
		
		# Register cleanup function
		self.channel.registerCleanup(DCDownloadSetupHandler.cleanup, (self.channel, self.block))

		# Setup file (create directories in its path if necessary)
		blockDir = dirname(self.block.path)
		if not exists(blockDir):
			try: makedirs(blockDir)
			except OSError:
				#TODO: notify user about this
				raise IOError
		
		f = self.channel.file = file(self.block.path, 'a+b')
		f.seek(0, 2)
		if self.channel._sizeRemaining != self.block.size - f.tell() + self.rollback:
			raise IOError

		# Notify xferWorker
		self.channel.notify()

		# Setup handler
		if self.rollback > 0:
			self.channel.setHandler(DCDownloadRollbackHandler(self.block, self.rollback))
		else:
			self.channel.setHandler(DCDownloadTransferHandler(self.block))
		
		return True
	
	def cleanup(pair):
		channel, block = pair
		channel.file = None
		channel.xferClosed()
		channel.item.setStatusWaiting(channel, block)
	
	cleanup = staticmethod(cleanup)

class DCDownloadRollbackHandler(ChannelHandler):
	def __init__(self, block, rollback):
		ChannelHandler.__init__(self)
		self.block = block
		self.rollback = rollback

	def onAttach(self, xfer):
		ChannelHandler.onAttach(self, xfer)
		self.file = xfer.file
		self.timestamp = time()
		self.speed = xfer._speed
		
		if xfer._compressed:
			self.zlib = decompressobj()
		else:
			self.zlib = None
		
		self.file.seek(-self.rollback, 1)
		self.speed.init(self.timestamp)
	
	def prepoll(self):
		self.timestamp = time()
		self.channel.setWindow(int(min(self.rollback, self.speed.getBandwidth(self.timestamp))))
	
	def poll(self):
		data = self.channel.sock.recv(self.rollback)
		netLen = len(data)
		
		if self.zlib:
			data = self.zlib.decompress(data, self.rollback)
			dataLen = len(data)
		else:
			dataLen = netLen
		
		self.rollback -= dataLen
		self.channel._sizeRemaining -= dataLen
		self.speed.usedBandwidth(self.timestamp, netLen, dataLen)

		try:
			localData = self.channel.file.read(dataLen)
			if cmp(data, localData) != 0:
				item = self.channel.getItem()
				item.removeSource(self.channel.getRemoteNick(), False, item.CAUSE_CORRUPTION)
				raise IOError

		except IOError:
			self.channel._item.setStatusWaiting(self.channel, self.block)
			self.channel.nextItem()

		if self.rollback <= 0:
			self.channel.setHandler(DCDownloadTransferHandler(self.block, False, self.zlib))

class DCDownloadTransferHandler(ChannelHandler):
	SPEEDSTAT_TIMEOUT = 3

	def __init__(self, block, initSpeed=True, zlib=None):
		ChannelHandler.__init__(self)
	
		self.block = block
		self.initSpeed = initSpeed
		self.window = 0
		self.zlib = zlib
	
	def onAttach(self, channel):
		ChannelHandler.onAttach(self, channel)

		self.file = self.channel.file
		self.sock = self.channel.sock
		self.speed = self.channel._speed
		self.timestamp = time()
		self.statStamp = self.timestamp
		
		if self.zlib == None:
			if channel._compressed:
				self.zlib = decompressobj()
		else:
			data = self.zlib.decompress(self.zlib.unconsumed_tail)
			self.channel._sizeRemaining -= len(data)
			self.file.write(data)
		
		if self.initSpeed:
			self.speed.init(self.timestamp)
		del self.initSpeed
	
	def prepoll(self):
		self.timestamp = time()
		self.channel.setWindow(int(self.speed.getBandwidth(self.timestamp)))
	
	def poll(self):
		if self.zlib:
			data = self.sock.recv(self.channel._sizeRemaining+100)
			netLen = len(data)
			data = self.zlib.decompress(data, self.channel._sizeRemaining)
			dataLen = len(data)
		else:
			data = self.sock.recv(self.channel._sizeRemaining)
			netLen = dataLen = len(data)
		
		self.file.write(data)
		self.channel._sizeRemaining -= dataLen
		self.speed.usedBandwidth(self.timestamp, netLen, dataLen)

		if self.timestamp - self.statStamp > self.SPEEDSTAT_TIMEOUT:
			self.channel._progress = (1 - float(self.channel._sizeRemaining) / self.block.size) * 100
			self.block.progress = self.channel._progress

			self.statStamp = self.timestamp
			DCWorker.getWorker().getXferWorker().onXferUpdate(self)

		if self.channel._sizeRemaining <= 0:
			if self.zlib:
				self.sock.unrecv(self.zlib.unconsumed_tail)
			
			self.channel._item.setStatusCompleted(self.channel, self.block)
			self.channel.nextItem()
