# -*- coding: ISO-8859-1 -*-
# This file is part of pyDC software
# Copyright 2002-2004 Anakim Border <aborder@users.sourceforge.net>
#
# pyDC is released under the terms of GPL licence.

from copy import copy
from time import localtime
from Tokenizer import Tokenizer
from weakref import proxy

from DCChallenge import *
from DCChannel import *
from DCSearch import DCSearch
from DCUser import *
from DCXfer import DCXfer
import DCWorker

class DCHub(DCChannel):
	MAX_LOG_ROWS = 1000

	LOG_STATUS = 1
	LOG_CHAT = 2
	LOG_PRIVCHAT = 3
	LOG_ERROR = 4

	NOTIFY_NEW = 1
	NOTIFY_INFO = 2
	NOTIFY_REMOVE = 3

	FOLLOW_REDIRECT = True
	RECONNECTION_DELAY = 2.5

	def __init__(self, addr, nick = None, password = None):
		DCChannel.__init__(self)
		self.addInterface('DCHub')

		# Local user info
		if nick:
			self._nick = nick
		else:
			self._nick = DCWorker.getWorker().getSettings().nick
		
		self._isOp = False
		self._isRegistered = False
		self.askingPwd = False
		self._password = password
		self.speedByte = DCUser.STATUS_DEFAULT

		# Hub data
		self._name = None
		self._logged = False
		self.reconnDelay = self.RECONNECTION_DELAY
		self._users = {}
		self._size = 0L
		self._log = []
		self.userEvents = {}
		
		# Final setup
		self.setHandler(DCHubLoginHandler())
		self.connect(addr, 411)
		self.appendLogRow(self.LOG_STATUS, 'Connecting to %s...' % self.getName())
	
	# Connection
	def reconnect(self, nodelay = False):
		if nodelay:
			self.reconnDelay = self.RECONNECTION_DELAY
			delay = 0
			self.appendLogRow(self.LOG_STATUS, 'Reconnecting...')
		else:
			self.reconnDelay *= 2
			delay = self.reconnDelay
			self.appendLogRow(self.LOG_STATUS, 'Reconnecting in %s seconds.' % int(self.reconnDelay))
		
		self.setHandler(DCHubLoginHandler())
		self.connect(delay = delay)
	
	def redirect(self, addr):
		if not self.FOLLOW_REDIRECT:
			self.appendLogRow(self.LOG_ERROR, 'Ignoring redirect.')
			#TODO: close channel
		else:
			self.appendLogRow(self.LOG_STATUS, 'Redirected to %s.' % addr)

			self.reconnDelay = self.RECONNECTION_DELAY
			self.setHandler(DCHubLoginHandler())
			self.connect(addr, 411)
			
	def stop(self):
		self.reconnDelay = None
		DCChannel.stop(self)

	def requireConnection(self, nick, forceActive = False):
		try:
			self.opLock.acquire()
			if not self._users.has_key(nick):
				return
		finally:
			self.opLock.release()

		worker = DCWorker.getWorker()
		worker.registerXferNick(self._nick, nick)
		
		if worker.getSettings().active:
			arg = '%s %s:%d' % (nick, worker.getLocalAddress(), worker.getSettings().port)
			self.sendCmd('ConnectToMe', arg)
		else:
			if not forceActive: self.sendCmd('RevConnectToMe', self._nick, nick)
	
	# Password
	def askPassword(self):
		if self._password != None:
			return self.setPassword(self._password)
		elif self.askingPwd:
			return
	
		self.lock.acquire()
		self.askingPwd = True
		for listener in self.getListeners('DCHub'):
			listener.onUserPassword(self)
		self.lock.release()

	def setPassword(self, password):
		self.askingPwd = False
		if password.find('$') != -1 or password.find('|') != -1:
			raise ValueError

		if not self._logged:
			self._isRegistered = True
			self._password = password
			self.sendCmd('MyPass', password)
	
	def clearPassword(self):
		self._password = None

	# User info
	def buildInfo(self):
		return DCWorker.getWorker().getInfoMask() % (self._nick, chr(self.speedByte))

	def updateInfo(self, silent):
		info = self.buildInfo()
		if not silent: self.appendLogRow(self.LOG_STATUS, 'Updating user info...')
		self.sendCmd('MyINFO', info)

	# Searches
	def startSearch(self, search, active):
		if not self._logged: return
	
		worker = DCWorker.getWorker()
		if active:
			arg = '%s:%d ' % (worker.getLocalAddress(), worker.getSettings().port)
		else:
			arg = 'Hub:%s ' % self._nick

		limit, size = search.getSizeLimit()
		if limit != DCSearch.LIMIT_NONE:
			if limit == DCSearch.LIMIT_ATLEAST:
				arg += 'T?F?%d?1?' % size
			else:
				arg += 'T?T?%d?1?' % size
		else:
			arg += 'F?F?0?1?'
		
		arg += '$'.join(search.getPattern().split(' '))
		self.sendCmd('Search', arg)

	# Users
	def addUser(self, nick):
		#Thread unsafe; caller must lock self.opLock
		if self._users.has_key(nick):
			return

		user = DCUser()
		user.nick = nick
		user.hub = proxy(self)

		self._users[nick] = user
		self.addUserEvent(user, self.NOTIFY_NEW)

	def removeUser(self, nick):
		#Thread unsafe; caller must lock self.opLock
		try:
			user = self._users[nick]
			del self._users[nick]
			self.addUserEvent(user, self.NOTIFY_REMOVE)

			user.hub = None
			if user.share > 0:
				self._size -= user.share
		except KeyError:
			pass

	def setInfo(self, info):
		self.opLock.acquire()
		try:
			user = self._users[info.nick]
			if user.share > 0:
				self._size -= user.share
			
			user.share = info.share
			user.description = info.description
			user.connection = info.connection
			user.status = info.status
			user.email = info.email
			self.addUserEvent(user, self.NOTIFY_INFO)
		
		except KeyError:
			info.op = 0
			info.hub = proxy(self)
			self._users[info.nick] = info
			self.addUserEvent(info, self.NOTIFY_NEW)

		self._size += info.share
		self.opLock.release()

	def addUserEvent(self, user, event):
		#Thread unsafe; caller must lock self.opLock
		if event == self.NOTIFY_NEW:
			self.userEvents[user.nick] = (self.NOTIFY_NEW, user)
		elif event == self.NOTIFY_INFO:
			if not self.userEvents.has_key(user.nick):
				self.userEvents[user.nick] = (self.NOTIFY_INFO, user)
		else:
			if not self.userEvents.has_key(user.nick) or self.userEvents[user.nick][0] != self.NOTIFY_NEW:
				self.userEvents[user.nick] = (self.NOTIFY_REMOVE, user)
			else:
				del self.userEvents[user.nick]

	def notifyUserEvents(self):
		lnew, linfo, lrem = [], [], []

		self.opLock.acquire()
		for nick, event in self.userEvents.iteritems():
			code, user = event
			if code == self.NOTIFY_NEW:
				lnew.append(user)
			elif code == self.NOTIFY_INFO:
				linfo.append(user)
			elif code == self.NOTIFY_REMOVE:
				lrem.append(user)

		self.userEvents = {}
		self.opLock.release()

		n, i, r = len(lnew) > 0, len(linfo) > 0, len(lrem) > 0
		self.lock.acquire()
		for listener in self.getListeners('DCHub'):
			if r: listener.onUserDisconnections(self, lrem)
			if i: listener.onUserInfos(self, linfo)
			if n: listener.onNewUsers(self, lnew)
		self.lock.release()

	# Messages & log
	def sendMsg(self, msg, nick = None):
		if not self._logged:
			raise ValueError, 'You must be logged to send messages.'
		elif nick:
			if nick == self._nick:
				raise ValueError, "You can't send messages to yourself."
			else:
				try:
					self.opLock.acquire()
					if not self._users.has_key(nick):
						raise ValueError, "Unknown user '%s' (maybe he went offline)." % nick
				finally:
					self.opLock.release()
		
		if nick == None:
			self.sock.send('<%s> %s|' % (self._nick, self.escapeMsg(msg)))
		else:
			self.sendCmd('To:', nick, 'From:', self._nick, '$<%s>' % self._nick, self.escapeMsg(msg))
			
	def escapeMsg(self, msg):
		return msg.replace('$', '&#36;').replace('|', '&#124;').replace('\n', '\r\n')

	def unescapeMsg(self, msg):
		return msg.replace('&#36;', '$').replace('&#124;', '|').replace('\r\n', '\n')
	
	def appendLogRow(self, type, text, notify = 1):
		self.opLock.acquire()
		row = (type, localtime(), text + '\n')
		self._log.append(row)
		if len(self._log) > DCHub.MAX_LOG_ROWS:
			del self._log[0]
		self.opLock.release()

		self.lock.acquire()
		for listener in self.getListeners('DCHub'):
			listener.onLogUpdate(self, row)
		self.lock.release()

	# Listeners
	def registerListener(self, listener, interface = None):
		EventGenerator.registerListener(self, listener, interface)
		self.lock.acquire()

		self.opLock.acquire()

		# Reverse changes to the user list not yet notified
		users = copy(self._users)
		for i in self.userEvents.iteritems():
			if i[1][0] == self.NOTIFY_NEW:
				del users[i[0]]
			elif i[1][0] == self.NOTIFY_REMOVE:
				users[i[0]] = i[1][1]
				
		self.opLock.release()

		listener.onNewUsers(self, users.values())
		self.lock.release()

	# Properties
	def getName(self):
		name = self._name
		if name == None:
			name = '%s:%s' % self._addr
		return name

	def getAddress(self):
		return self._addr

	def getNick(self):
		return self._nick

	def isLogged(self):
		return self._logged

	def isOp(self):
		return self._isOp
	
	def isRegistered(self):
		return self._isRegistered

	def getUsers(self):
		return self._users.values()

	def getUserNum(self):
		return len(self._users)

	def getUserByNick(self, nick):
		self.opLock.acquire()
		try: user = self._users[nick]
		except KeyError: user = None
		self.opLock.release()
		return user

	def getSize(self):
		return self._size

	def getLogRows(self):
		return self._log

	#=================#
	# Private members #
	#=================#
	
	def setName(self, name):
		self._name = name
		self.lock.acquire()
		for listener in self.getListeners('DCHub'):
			listener.onHubInfo(self)
		self.lock.release()

	def onConnection(self):
		DCWorker.getWorker().refreshIP(self.sock.getsockname()[0], self)
		
		self.lock.acquire()
		for listener in self.getListeners('DCHub'):
			listener.onHubConnection(self)
		self.lock.release()

		self.appendLogRow(self.LOG_STATUS, 'Connection established.')

	def onDisconnection(self):
		DCChannel.onDisconnection(self)
		self.appendLogRow(self.LOG_STATUS, 'Connection closed.')
		
		if self._logged:
			self._logged = False
			DCWorker.getWorker().updateInfo()
		
		self._isOp = False
		self._isRegistered = False
		self._size = 0L
		self.userEvents = {}
		
		for user in self._users.itervalues():
			user.hub = None
		self._users = {}
		
		if self.reconnDelay:
			self.reconnect()
		
		self.lock.acquire()
		for listener in self.getListeners('DCHub'):
			listener.onHubDisconnection(self)
		self.lock.release()

	def onSocketError(self, e):
		self.appendLogRow(self.LOG_ERROR, 'Socket error: %s.' % e[1])
		
		self.lock.acquire()
		for listener in self.getListeners('DCHub'):
			listener.onConnectionError(self)
		self.lock.release()
		
		if self.reconnDelay:
			self.reconnect()

	name = property(getName)
	address = property(getAddress)
	nick = property(getNick)
	logged = property(isLogged)
	op = property(isOp)
	registered = property(isRegistered)
	users = property(getUsers)
	userNum = property(getUserNum)
	size = property(getSize)
	log = property(getLogRows)

class DCHubBaseHandler(DCChannelHandler):
	def __init__(self, commands):
		commands['HubName']   = (-1, DCHubBaseHandler.onHubname)
		commands['ForceMove'] = ( 1, DCHubBaseHandler.onForceMove)
		DCChannelHandler.__init__(self, commands)
		self.opLock = None
	
	def onAttach(self, hub):
		DCChannelHandler.onAttach(self, hub)
		self.opLock = hub.opLock
	
	def onChatMessage(self, message):
		self.channel.appendLogRow(self.channel.LOG_CHAT, self.channel.unescapeMsg(message))
	
	def onForceMove(self, address):
		self.channel.redirect(address)
	
	def onHubname(self, name):
		self.channel.setName(name)

class DCHubLoginHandler(DCHubBaseHandler):
	def __init__(self):
		cmds = { 'Lock' :           (2, DCHubLoginHandler.onLock),
		         'ValidateDenide' : (1, DCHubLoginHandler.onValidateDenide),
		         'GetPass' :        (0, DCHubLoginHandler.onGetPass),
		         'BadPass' :        (0, DCHubLoginHandler.onBadPass),
		         'LogedIn' :        (1, DCHubLoginHandler.onLogedIn),
		         'Hello' :          (1, DCHubLoginHandler.onHello) }
		DCHubBaseHandler.__init__(self, cmds)

	def onInvalidCommand(self, cmd):
		self.channel.appendLogRow(self.channel.LOG_ERROR, 'Invalid command \'%s\' received: %s' % (cmd.name, cmd.data))
		raise ValueError

	def onUnknownCommand(self, cmd):
		pass

	def onLock(self, key, pk):
		self.channel.sendCmd('Key', solveChallenge(key))
		self.channel.sendCmd('ValidateNick', self.channel._nick)

	def onValidateDenide(self, nick):
		self.channel.appendLogRow(self.channel.LOG_ERROR, 'Nick \'%s\' already in use; connection aborted.' % self.channel._nick)
		raise ValueError

	def onGetPass(self):
		self.channel.askPassword()

	def onBadPass(self):
		self.channel.appendLogRow(self.channel.LOG_ERROR, 'Wrong password; connection aborted.')
		self.channel.clearPassword()
		self.channel.reconnDelay = None
		raise ValueError

	def onLogedIn(self, nick):
		if nick != self.channel._nick: raise ValueError
		self.channel._isOp = True

	def onHello(self, nick):
		if nick != self.channel._nick:
			return

		self.channel._logged = True
		self.channel.reconnDelay = self.channel.RECONNECTION_DELAY
		self.channel.setHandler(DCHubHandler())
		self.channel.sendCmd('Version', '1.0091')
		DCWorker.getWorker().updateInfo(True)
		self.channel.sendCmd('GetNickList')

class DCHubHandler(DCHubBaseHandler):
	def __init__(self):
		cmds = { 'NickList' :       (-1, DCHubHandler.onNickList),
		         'OpList' :         (-1, DCHubHandler.onOpList),
		         'Hello' :          ( 1, DCHubHandler.onHello),
		         'Quit' :           ( 1, DCHubHandler.onQuit),
		         'To:' :            (-1, DCHubHandler.onTo),
		         'MyINFO' :         (-1, DCHubHandler.onMyInfo),
		         'ConnectToMe' :    ( 2, DCHubHandler.onConnectToMe),
		         'RevConnectToMe' : ( 2, DCHubHandler.onRevConnectToMe),
		         'Search' :         (-1, DCHubHandler.onSearch),
		         'SR' :             (-1, DCHubHandler.onSR) }
		DCHubBaseHandler.__init__(self, cmds)
	
	# Commands
	def poll(self):
		DCHubBaseHandler.poll(self)
		self.channel.notifyUserEvents()
	
	def onInvalidCommand(self, cmd):
		if cmd.name in ('Search'): return
		self.channel.appendLogRow(self.channel.LOG_ERROR, 'Invalid command \'%s\' received: %s' % (cmd.name, cmd.data))
	
	def onUnknownCommand(self, cmd):
		if cmd.name in ('MultiSearch'): return
		self.channel.appendLogRow(self.channel.LOG_ERROR, 'Unknown command \'%s\' received.' % cmd.name)
	
	def onNickList(self, nickList):
		try:
			self.opLock.acquire()
			nicks = nickList.split('$$')
			if len(nicks) < 1: raise InvalidCommand

			for nick in nicks[:-1]:
				self.channel.addUser(nick)
				self.channel.sendCmd('GetINFO', nick, self.channel.getNick())
		finally:
			self.opLock.release()

	def onOpList(self, nickList):
		try:
			self.opLock.acquire()
			nicks = nickList.split('$$')
			if len(nicks) < 1: raise InvalidCommand

			for nick in nicks[:-1]:
				try: self.channel._users[nick].op = 1
				except KeyError: continue
		finally:
			self.opLock.release()

	def onHello(self, nick):
		self.opLock.acquire()
		self.channel.addUser(nick)
		self.opLock.release()

	def onQuit(self, nick):
		self.opLock.acquire()
		self.channel.removeUser(nick)
		self.opLock.release()

	def onTo(self, data):
		try:
			t = Tokenizer(data)
			
			dest = t.next('From:').strip()
			if dest != self.channel._nick: raise InvalidCommand, "malformed message or it's not for me"
			source = t.next('$').strip()
			
			data = t.flush()
			if len(data) > 0:
				self.channel.appendLogRow(self.channel.LOG_PRIVCHAT, '<%s> %s' % (source, self.channel.unescapeMsg(data)))
		
		except ValueError:
			raise InvalidCommand
	
	def onMyInfo(self, data):
		try:
			t = Tokenizer(data)
		
			dest = t.next()
			if dest != '$ALL': raise ValueError

			# Nick
			info = DCUser()
			info.nick = t.next()

			# Description
			info.description = t.next('$ $')
			if len(info.description) == 0:
				info.description = None

			# Status flag
			conn = t.next('$')
			info.status = conn[-1]
			conn = conn[:-1]

			# Connection
			connTable = { '28.8Kbps'   : DCUser.CONN_28K,
			              '33.6Kbps'   : DCUser.CONN_33K,
			              '56Kbps'     : DCUser.CONN_56K,
			              'Modem'      : DCUser.CONN_MODEM,
			              'ISDN'       : DCUser.CONN_ISDN,
			              'Wireless'   : DCUser.CONN_WIRELESS,
			              'Microwave'  : DCUser.CONN_MICROWAVE,
			              'Satellite'  : DCUser.CONN_SATELLITE,
			              'DSL'        : DCUser.CONN_DSL,
			              'Cable'      : DCUser.CONN_CABLE,
			              'LAN(T1)'    : DCUser.CONN_T1,
			              'LAN(T3)'    : DCUser.CONN_T3 }
			try: info.connection = connTable[conn]
			except KeyError:
				# DCGUI uses localized connection IDs
				# Since we can't list them all, ignore
				# errors and mark connection as CONN_NONE
				info.connection = DCUser.CONN_NONE

			# Email
			info.email = t.next('$')
			if len(info.email) == 0:
				info.email = None
		
		except ValueError:
			raise InvalidCommand
		except KeyError:
			raise InvalidCommand
		
		# Share
		try:
			info.share = long(t.next('$'))
		except ValueError:
			# Bots don't use that field, so set share to 0
			info.share = 0
		
		if info.share < 0:
			raise InvalidCommand
			
		if t.hasMore():
			raise InvalidCommand
		
		self.channel.setInfo(info)

	def onConnectToMe(self, nick, addr):
		if nick != self.channel.getNick():
			raise InvalidCommand

		parts = addr.split(':')
		try: addr = (parts[0], int(parts[1]))
		except ValueError: raise InvalidCommand

		DCWorker.getWorker().addJob(DCXfer(False, addr, nick))

	def onRevConnectToMe(self, nick, myNick):	
		if myNick != self.channel._nick:
			raise InvalidCommand
		self.channel.requireConnection(nick, True)

	def onSearch(self, data):
		try:
			t = Tokenizer(data)
		
			# Search destination
			dest = t.next(':')
			if dest == 'Hub':
				activeSearch = False
				dest = t.next()
				# TODO: add nick check
			else:
				activeSearch = True
				destPort = int(t.next())
				if destPort < 0 or destPort >= 0xffff: raise InvalidCommand

			# Search params
			limit = t.next('?')
			ismin = t.next('?')
			if limit == 'F':
				limit = DCSearch.LIMIT_NONE
			elif limit == 'T':
				if ismin == 'T':
					limit = DCSearch.LIMIT_ATLEAST
				elif ismin == 'F':
					limit = DCSearch.LIMIT_ATMOST
				else:
					raise InvalidCommand
			else:
				raise InvalidCommand

			size = long(t.next('?'))
			if size < 0 or (limit == DCSearch.LIMIT_NONE and size > 0):
				# NOTE: Too many broken clients: simply ignore this
				return

			tp = int(t.next('?'))
			if tp < 1 or tp > 8:
				raise InvalidCommand

			pattern = t.flush()
			if pattern == None:
				raise InvalidCommand
			patter = pattern.replace('$', ' ')

			# Perform search
			worker = DCWorker.getWorker()
			slots = [max(worker.getSettings().getSlotNum() - worker.getUsedSlots(), 0), worker.getSettings().getSlotNum()]
			hubAddr = self.channel.sock.getpeername()
			if hubAddr == None: raise ValueError

			for r in worker.getLocalList().search(pattern, (limit, size), 4):
				data =  '$SR %s %s\5%d %d/%d\5%s (%s:%d)' % (self.channel._nick, r[0], r[1], slots[0], slots[1], self.channel._name, hubAddr[0], hubAddr[1])
				if activeSearch:
					worker.getSearchWorker().sendResult((dest, destPort), data + '|')
				else:
					data += '\5%s|' % dest
					self.channel.sock.send(data)
		
		except ValueError:
			raise InvalidCommand

	def onSR(self, data):
		DCWorker.getWorker().getSearchWorker().passiveResult(data)
