# This file is part of pyDC software
# Copyright 2002-2004 Anakim Border <aborder@users.sourceforge.net>
#
# pyDC is released under the terms of GPL licence.

from bloom import BloomSummary
from copy import copy
from os import listdir, sep, stat
from os.path import abspath, basename, isdir

from DCSearch import DCSearch

class DCLocalList:
	def __init__(self):
		self._list = []
		self._paths = []
		self._size = 0
		self._itemCount = 0
		self.sizeTbl = SizeTable()
		self.bloomSummary = None
	
	def scan(self, path):
		path = abspath(path)
		if not isdir(path):
			raise ValueError
		
		root = (basename(path), -1, None, [])
		self.__scan(path, root, root[3])
		self._list.append(root)
		self._paths.append(path)
	
	def __scan(self, path, parent, list):
		items = listdir(path)
		for i in items:
			try:
				fullpath = path + sep + i
				if isdir(fullpath):
					dir = (i, -1, parent, [])
					self.__scan(fullpath, dir, dir[3])
					
					if len(dir[3]):
						list.append(dir)
				else:
					size = stat(fullpath).st_size
					file = (i, size, parent, None)
					
					list.append(file)
					self.sizeTbl.add(file)
					self._size += size
				
				self._itemCount += 1
			except OSError:
				pass
	
	def freeze(self):
		self.sizeTbl.freeze()
		
		self.bloomSummary = BloomSummary(self._itemCount)
		dlist = copy(self._list)
		while len(dlist):
			for i in dlist[0][3]:
				self.__storePattern(i[0])
				if i[1] == -1: #dir
					dlist.append(i)
			
			del dlist[0]
	
	def __storePattern(self, name):
		pbs = []
		s = ''
		for c in name:
			if c.isalnum():
				s += c
			else:
				if len(s) > 0:
					pbs.append(s.lower())
					s = ''
		if len(s): pbs.append(s.lower())
		
		for p in pbs:
			self.bloomSummary.add(p)
	
	def search(self, pattern, limit = None, maxnum = 0xffff):
		aux = [ (len(b), b) for b in pattern.lower().split('$') if len(b) ]
		aux.sort()
		aux.reverse()
		
		auxLen = len(aux)
		if auxLen == 0 or aux[0][0] < 4:
			return []
		pbs = [ b[1] for b in aux ]
		
		for b in pbs:
			if not self.bloomSummary.match(b):
				return []
		
		matches, res = [], []
		
		if limit and limit[0] != DCSearch.LIMIT_NONE:
			self.sizeTbl.search(pbs, limit[0], limit[1], matches, maxnum)
		else:
			for root in self._list:
				self.__unsizedSearch(root, pbs, matches, maxnum)
				if len(matches) == maxnum: break
			
		for match in matches:
			path = match[0]
			size = match[1]
			while match[2]:
				match = match[2]
				path = '%s\%s' % (match[0], path)
			
			res.append((path, size))
		
		return res
		
	def __unsizedSearch(self, dir, pbs, matches, maxnum):
		pbsLen = len(pbs)
		
		for item in dir[3]:
			if item[3] == None: #file
				pbsCount = 0
				name = item[0].lower()
				for b in pbs:
					if name.find(b) != -1:
						pbsCount += 1
				
				if pbsCount == pbsLen:
					matches.append(item)
					if len(matches) == maxnum:
						return
			
			else:
				aux, auxCount = [], 0
				name = item[0].lower()
				for p, b in enumerate(pbs):
					if name.find(b) != -1:
						aux.append(p)
						auxCount += 1
				
				if auxCount > 0:
					localPbs = copy(pbs)
					aux.reverse()
					for p in aux:
						del localPbs[p]
				else:
					localPbs = pbs
				
				self.__unsizedSearch(item, localPbs, matches, maxnum)
				if len(matches) == maxnum:
					return
	
	def getPaths(self):
		return copy(self._paths)
	
	def getList(self):
		return copy(self._list)

	def getSize(self):
		return self._size

	paths = property(getPaths)
	list = property(getList)
	size = property(getSize)

class SizeTable:
	def __init__(self):
		self.buckets = [ [[], 0xffffffffffL, 0] ]
		self.bucketNum = 1
		self.bucketSize = 0
	
	def add(self, file):
		b = self.buckets[0]
		b[0].append(file)
		
		size = file[1]
		if b[1] > size:
			b[1] = file[1]
		elif b[2] < size:
			b[2] = size
	
	def freeze(self):
		self.bucketSize = min(len(self.buckets[0][0]) / 100, 100)
	
		while True:
			completed = True
			for p, b in enumerate(self.buckets):
				if len(b[0]) > self.bucketSize:
					med = b[1] + (b[2]-b[1])/2
					if med == b[1]: continue
					
					l, h = [], []
					for i in b[0]:
						if i[1] < med:
							l.append(i)
						else:
							h.append(i)
					
					if len(l) == 0:
						b[1] = med
					elif len(h) == 0:
						b[2] = med
					else:
						self.buckets.insert(p+1, [h, med, b[2]])
						b[0] = l
						b[2] = med
						self.bucketNum += 1
				
					completed = False
					break
		
			if completed: break
	
	def search(self, pbs, limit, size, matches, maxnum):
		if limit == DCSearch.LIMIT_ATLEAST:
			self.__searchBigger(pbs, size, matches, maxnum)
		else:
			self.__searchSmaller(pbs, size, matches, maxnum)
	
	def __searchSmaller(self, pbs, size, matches, maxnum):
		idx = self.bucketNum-1
		while self.buckets[idx][1] > size and idx >= 0:
			idx -= 1
		if idx == -1: return
		
		for file in self.buckets[idx][0]:
			if file[1] > size:
				continue
			
			name = file[0].lower()
			found = True
			for b in pbs:
				if name.find(b) == -1:
					found = False
					break
			
			if found:
				matches.append(file)
				if len(matches) == maxnum:
					return
		
		while idx > 0:
			idx -= 1
			for file in self.buckets[idx][0]:
				name = file[0].lower()
				found = True
				for b in pbs:
					if name.find(b) == -1:
						found = False
						break
				
				if found:
					matches.append(file)
					if len(matches) == maxnum:
						return
	
	def __searchBigger(self, pbs, size, matches, maxnum):
		idx = 0
		while idx < self.bucketNum and self.buckets[idx][2] < size:
			idx += 1
		if idx == self.bucketNum: return
		
		for file in self.buckets[idx][0]:
			if file[1] < size:
				continue
			
			name = file[0].lower()
			found = True
			for b in pbs:
				if name.find(b) == -1:
					found = False
					break
			
			if found:
				matches.append(file)
				if len(matches) == maxnum:
					return
		
		while idx < self.bucketNum-1:
			idx += 1
			for file in self.buckets[idx][0]:
				name = file[0].lower()
				found = True
				for b in pbs:
					if name.find(b) == -1:
						found = False
						break
				
				if found:
					matches.append(file)
					if len(matches) == maxnum:
						return
