# This file is part of pyDC software
# Copyright 2002-2004 Anakim Border <aborder@users.sourceforge.net>
#
# pyDC is released under the terms of GPL licence.

from copy import copy
from random import randint
from threading import Lock
from time import time
import pickle

from EventGenerator import *
from DCQueueItem import *
from DCSourceSearch import *
from DCXfer import *
from Job import *

class DCQueue(object, Job, EventGenerator):
	RETRY_TIME = 30

	SAVE_DELAY = 15
	QUEUE_FILE = 'queue.pk'

	SEARCH_TRIGGER  = 3000
	SEARCH_INTERVAL =   60
	
	BLOCK_SIZE = 20*1024*1024
	MAX_CONCURRENT_BLOCKS = 1
	AVOID_OLD_CLIENTS = False

	def __init__(self, worker):
		EventGenerator.__init__(self)
		self.worker = worker

		self._items = []
		self.running = True
		self.saveTime = 0
		self.searchTime = 0
		self.opLock = Lock()

	def addItem(self, item, path = None):
		queueItem = DCQueueItem(item, path)
		self.addQueueItem(queueItem)
		return queueItem

	def removeItem(self, item):
		self.opLock.acquire()
		try:
			idx = self._items.index(item)
			self.worker.getXferWorker().stopXfer(item)
			del self._items[idx]
			
			removed = True
			self.saveTime = time() + self.SAVE_DELAY
		except ValueError:
			removed = False
		self.opLock.release()
		
		if removed:
			self.lock.acquire()
			for listener in self.getListeners():
				listener.onItemRemoved(self, item)
			self.lock.release()
	
	def removeCompletedItems(self):
		self.opLock.acquire()
		aux = [ i for i in self._items if i.getStatus() == DCQueueItem.STATUS_COMPLETED ]
		self.opLock.release()
		
		for item in aux:
			self.removeItem(item)
			
	def removeAllItems(self):
		for item in self.getItems():
			self.removeItem(item)

	def getItems(self):
		self.opLock.acquire()
		items = copy(self._items)
		self.opLock.release()

		return items
	
	def getItemByNick(self, nick):
		try:
			self.opLock.acquire()
			for item in self._items:
				if not item.needsDownload(): continue
				for source in item.getSources():
					if source.getNick() == nick:
						return item, source
		
			return None, None
		finally:
			self.opLock.release()

	def load(self):
		try:
			self.opLock.acquire()
			self._items = []
			for item in pickle.load(file(self.QUEUE_FILE, 'rb')):
				self._items.append(DCQueueItem(item))
		finally:
			self.opLock.release()

	def save(self):
		try:
			self.opLock.acquire()
			aux = []
			for item in self._items:
				serialized = item.serialize()
				if serialized: aux.append(item.serialize())
			
			pickle.dump(aux, file(self.QUEUE_FILE, 'wb'))
			self.saveTime = 0
		finally:
			self.opLock.release()

	def poll(self):
		if self.running:
			requests, alternates = [], []
			t = time()
			canSearch = self.SEARCH_TRIGGER > 0 and t - self.searchTime > self.SEARCH_INTERVAL
			lockedSources = [ x.getRemoteNick() for x in self.worker.getJobs() if isinstance(x, DCXfer) and x.getType() == DCXfer.TYPE_DOWNLOAD ]

			self.opLock.acquire()
			for item in self._items[:100]:
				if not item.needsDownload(): continue
				
				if t - item.timestamp > DCQueue.RETRY_TIME:
					requests.append(item)
			
				if canSearch and isinstance(item, DCQueueItem):
					if item._retries > self.SEARCH_TRIGGER:
						alternates.append(item)
					elif item._retries != -1:
						item._retries += 1
			self.opLock.release()
			
			for item in requests:
				item.timestamp = t
				item.sendRequest(lockedSources)
			
			n = len(alternates)
			if n > 0:
				item = alternates[randint(0, n-1)]
				DCSourceSearch(item).start()
				self.searchTime = t
				item._retries = 0
			
			if self.saveTime > 0 and self.saveTime < t:
				self.save()
				
			return 0

		else:
			return 1

	def stop(self):
		if self.saveTime > 0: self.save()
		self.running = False

	def isAlive(self):
		return self.running
		
	def onItemStatus(self, item):
		self.opLock.acquire()
		try: idx = self._items.index(item)
		except: idx = -1
		self.opLock.release()
		
		if idx >= 0:
			self.opLock.acquire()
			del self._items[idx]
			
			if item.getStatus() in (DCQueueItem.STATUS_COMPLETED, DCQueueItem.STATUS_ERROR):
				self._items.append(item)
			else:
				self._items.insert(0, item)
			self.opLock.release()
		
			self.lock.acquire()
			for listener in self.getListeners():
				listener.onItemStatus(self, item)
			self.lock.release()
	
	def onItemSourceChange(self, item):
		self.saveTime = time() + self.SAVE_DELAY
	
	###################
	# Private members #
	###################
	
	def addQueueItem(self, item, priority = False):
		try:
			self.opLock.acquire()
			for i in self._items:
				if i.getPath() == item.getPath():
					raise ItemAlreadyAddedError

			if priority:
				self._items = [ item ] + self._items
			else:
				self._items.append(item)
			
			self.saveTime = time() + self.SAVE_DELAY
		finally:
			self.opLock.release()

		self.lock.acquire()
		for listener in self.getListeners():
			listener.onNewItem(self, item)
		self.lock.release()
	
	items = property(getItems, None, None, None)

class ItemAlreadyAddedError(Exception): pass
