# This file is part of pyDC software
# Copyright 2002-2004 Anakim Border <aborder@users.sourceforge.net>
#
# pyDC is released under the terms of GPL licence.

from copy import copy
from os import open, close, read, remove, rename, sep, stat, write, O_APPEND, O_RDONLY, O_WRONLY
from os.path import abspath
from thread import start_new_thread
from threading import Lock
import random

from DCItem import DCItem
import DCWorker

try:
	from os import O_BINARY
except:
	O_BINARY = 0

class DCQueueItem(object):
	STATUS_WAITING = 0
	STATUS_RUNNING = 1
	STATUS_COMPLETED = 2
	STATUS_ERROR = 3
	
	STATUS_LABELS = { STATUS_WAITING   : 'Waiting',
	                  STATUS_RUNNING   : 'Running',
	                  STATUS_COMPLETED : 'Completed',
	                  STATUS_ERROR     : 'Error' }
	CORRUPTED_LABEL = 'Corrupted'
	
	CAUSE_XFER_ERROR = 1
	CAUSE_CORRUPTION = 2
	
	def __init__(self, item, path = None):
		if isinstance(item, DCItem):
			base = DCWorker.getWorker().getSettings().getDownloadDir()
			if path:
				self._path = base + sep + path
			else:
				self._path = base + sep + item.path[item.path.rindex('\\')+1:]
			self._path = abspath(self._path)
			
			if len(self._path) < len(base):
				raise IllegalLocalPathError
			
			if self._path[:len(base)] != base:
				raise IllegalLocalPathError
			
			self._sources = [ DCQueueItemSource(item.userNick, item.path) ]
			self._status = self.STATUS_WAITING
			self._blocks = []
			self.blockSerial = 1
			
		elif isinstance(item, DCSerializedQueueItem):
			self._path = item.path
			self._sources = item.sources
			self._status = item.status
			
			self._blocks = []
			for b in item.blocks:
				self._blocks.append(DCQueueItemBlock(*b))
			
			self.blockSerial = item.blockSerial
			
		else:
			raise IllegalItem
		
		self._size = item.size
		self._retries = 0
		self.timestamp = 0
		self.lock = Lock()
		self.joinLock = Lock()

	def addSource(self, item):
		if self._size != item.size:
			raise ItemNotMatching

		try:
			self.lock.acquire()
			
			for source in self._sources:
				if source.nick == item.userNick and source.path == item.path:
					raise ExistingSourceError
	
			self._sources.append(DCQueueItemSource(item.userNick, item.path))
			self._timestamp = 0
	
			statusUpd = False
			if self._status == DCQueueItem.STATUS_ERROR:
				self._status = DCQueueItem.STATUS_WAITING
				statusUpd = True
		finally:
			self.lock.release()

		DCWorker.getWorker().getQueue().onItemSourceChange(self)
		if statusUpd: DCWorker.getWorker().getQueue().onItemStatus(self)

	def removeSource(self, nick, stop = True, cause = 0):
		self.lock.acquire()
		source = None
		for p, s in enumerate(self._sources):
			if s.getNick() == nick:
				source = s
				del self._sources[p]
				break
		self.lock.release()
		
		if source == None:
			return
		
		if cause == self.CAUSE_CORRUPTION:
			self._retries = -1 #signal possible corruption
		
		worker = DCWorker.getWorker()
		if stop:
			worker.getXferWorker().stopXfer(self, source)
		
		worker.getQueue().onItemSourceChange(self)
		if len(self._sources) == 0:
			self._status = self.STATUS_ERROR
			worker.getQueue().onItemStatus(self)
	
	def serialize(self):
		self.lock.acquire()
		progress = [ b.progress for b in self._blocks ]
		data = DCSerializedQueueItem(self._path, self._size, self._sources, self._status, self._blocks, progress, self.blockSerial)
		self.lock.release()
		return data

	def getPath(self):
		return self._path
	
	def getSize(self):
		return self._size
	
	def setSize(self, size):
		try:
			self.lock.acquire()
			
			if self._size != -1: return
			self._size = size
			if len(self._blocks):
				self._blocks[0].size = size
		finally:
			self.lock.release()
	
	def getSources(self):
		self.lock.acquire()
		sources = copy(self._sources)
		self.lock.release()
		return sources
	
	def getOnlineSources(self):
		self.lock.acquire()
		nicks = [ s._nick for s in self._sources ]
		self.lock.release()
		
		aux = []
		for nick in nicks:
			for user in DCWorker.getWorker().getUsersByNick(nick):
				aux.append(user)
		return aux
	
	def getStatus(self):
		return self._status
	
	def getStatusLabel(self):
		if self._retries == -1 and self._status != self.STATUS_ERROR:
			return self.CORRUPTED_LABEL
		else:
			return self.STATUS_LABELS[self._status]
	
	def getProgress(self):
		self.lock.acquire()
		downloaded100 = 0.0
		for block in self._blocks:
			downloaded100 += block.progress * block.size
		self.lock.release()
		return max(0.0, downloaded100 / self._size)
	
# 	def getXfers(self):
# 		self.lock.acquire()
# 		xfers = []
# 		for xfer in self._xfers:
# 			if xfer._status == self.STATUS_RUNNING:
# 				xfers.append(xfer._xfer)
# 		self.lock.release()
# 
# 		return xfers

	def getPattern(self):
		try:
			self.lock.acquire()
			l = len(self._sources)
			if l == 0: return None
			source = self._sources[random.randint(0, l-1)]
		finally:
			self.lock.release()

		pattern, path = '', source.getPath()
		path = path[path.rindex('\\')+1:]
		for c in path:
			if c.isalnum():
				pattern += c
			else:
				pattern += ' '

		return pattern.strip()
	
	def needsDownload(self):
		self.lock.acquire()
		needs = False
		
		if self._status == self.STATUS_WAITING:
			needs = True
		
		elif self._status == self.STATUS_RUNNING:
			waiting, running, counter = 0, 0, 0
			for block in self._blocks:
				counter += block.size
				if block.status == self.STATUS_RUNNING:
					running += 1
				elif block.status == self.STATUS_WAITING:
					waiting += 1
			
			if running < DCWorker.getWorker().getQueue().MAX_CONCURRENT_BLOCKS and (waiting > 0 or counter < self._size):
				needs = True
		
		self.lock.release()
		
		return needs

	def getBlock(self, xfer, forceSingle = False):
		try:
			self.lock.acquire()
			
			if self._size == -1 or forceSingle:
				if len(self._blocks) == 0:
					self._blocks.append(DCQueueItemBlock(self._path, 0, self._size))
				return self._blocks[0]
			
			if xfer.supportsSegments():
				start = 0
				for pos, b in enumerate(self._blocks):
					if start < b.offset:
						name, size = self.getBlockInfo(start, b.offset - start)
						block = DCQueueItemBlock(name, start, size)
						self._blocks.insert(pos, block)
						return block
					elif b.status == self.STATUS_WAITING:
						try:
							size = stat(b.path).st_size
						except OSError:
							size = 0
						
						defaultSize = DCWorker.getWorker().getQueue().BLOCK_SIZE
						if defaultSize > 0:
							oldSize = b.size
							b.size = min(b.size, (size / defaultSize + 1) * defaultSize)
							b.progress = b.progress * oldSize / b.size
						
						return b
					else:
						start += b.size
				
				if start == self._size:
					return None
				
				name, size = self.getBlockInfo(start, self._size - start)
				block = DCQueueItemBlock(name, start, size)
				self._blocks.append(block)
				return block
			else:
				if len(self._blocks) == 0:
					block = DCQueueItemBlock(self._path, 0, self._size)
					self._blocks.append(block)
				else:
					block = self._blocks[-1]
					if block.status == self.STATUS_RUNNING:
						return None
					
					size = self._size - block.offset
					if block.size == size and block.status == self.STATUS_COMPLETED:
						return None
					
					block.status = self.STATUS_WAITING
					block.progress = block.progress * block.size / size
					block.size = size
				
				return block
		
		finally:
			self.lock.release()
	
	def hasManyBlocks(self):
		try:
			self.lock.acquire()
			num = len(self._blocks)
			if num > 1:
				return True
			else:
				size = sum([ b.size for b in self._blocks ])
				return size < self._size
		finally:
			self.lock.release()
	
	def getBlockPaths(self):
		self.lock.acquire()
		paths = [ b.path for b in self._blocks ]
		self.lock.release()
		return paths

	def setStatusRunning(self, source, block):
		self.lock.acquire()
		
		if block.status != self.STATUS_WAITING:
			retval, notify = False, False
		else:
			block.status = self.STATUS_RUNNING
			block.xfer = source
			self._retries = 0
			retval = True
			
			if self._status == self.STATUS_WAITING:
				self._status = self.STATUS_RUNNING
				notify = True
			else:
				notify = False
		
		self.lock.release()

		if notify: DCWorker.getWorker().getQueue().onItemStatus(self)
		return retval

	def setStatusWaiting(self, source, block):
		self.lock.acquire()
		
		retval = False
		if block.xfer == source:
			block.status = self.STATUS_WAITING
			block.xfer = None
			retval = True
			
			running = False
			for b in self._blocks:
				if b.status == self.STATUS_RUNNING:
					running = True
					break
			
			if not running:
				self._status = self.STATUS_WAITING
		
		self.lock.release()

		if retval: DCWorker.getWorker().getQueue().onItemStatus(self)
		return retval

	def setStatusCompleted(self, source, block):
		self.lock.acquire()
		retval = False
		joinList = []
		
		if block.xfer == None or block.xfer == source:
			block.status = self.STATUS_COMPLETED
			block.xfer = None
			block.progress = 100
			retval = True

			waiting, running = 0, 0
			counter = 0
			for b in self._blocks:
				counter += b.size
				if b.status == self.STATUS_RUNNING:
					running += 1
					break
				elif b.status == self.STATUS_WAITING:
					waiting += 1

			#TODO: what about STATUS_ERROR
			if running:
				self._status = self.STATUS_RUNNING
			elif waiting or counter < self._size:
				self._status = self.STATUS_WAITING
			else:
				self._status = self.STATUS_COMPLETED

		self.lock.release()

		if retval:
			DCWorker.getWorker().getQueue().onItemStatus(self)
			start_new_thread(self.joinBlocks, ())
		
		return retval
	
	def sendRequest(self, lockedSources):
		for user in [ s for s in self.getOnlineSources() if not s.getNick() in lockedSources ]:
			lockedSources.append(user.getNick())
			user.hub.requireConnection(user.getNick())

	###################
	# Private methods #
	###################
	
	def getBlockInfo(self, start, maxSize):
		# Thread unsafe: caller must acquire self.lock
		if start == 0:
			name = self._path
		else:
			blockNum = self.blockSerial
			self.blockSerial += 1
			name = '%s.block%d' % (self._path, blockNum)
		
		defaultSize = DCWorker.getWorker().getQueue().BLOCK_SIZE
		if defaultSize > 0:
			size = min(defaultSize, maxSize)
		else:
			size = maxSize
		
		return name, size
	
	def joinBlocks(self):
		try:
			self.joinLock.acquire()
			
			while True:
				self.lock.acquire()
				found = False
				for i in xrange(len(self._blocks)-1):
					a, b = self._blocks[i], self._blocks[i+1]
					if b.offset == a.offset + a.size and a.status == self.STATUS_COMPLETED and b.status == self.STATUS_COMPLETED:
						found = True
						break
				self.lock.release()
				
				if found:
					try:
						fa = open(a.path, O_WRONLY | O_APPEND | O_BINARY)
					except OSError:
						a.progress = 0
						self.setStatusWaiting(None, a)
						continue
					
					try:
						fb = open(b.path, O_RDONLY | O_BINARY)
					except OSError:
						close(fa)
						b.progress = 0
						self.setStatusWaiting(None, b)
						continue
					
					try:
						while True:
							data = read(fb, 16384)
							if len(data) == 0: break
							write(fa, data)
						
						close(fa)
						close(fb)
						remove(b.path)
						
						self.lock.acquire()
						a.size += b.size
						self._blocks.remove(b)
						self.lock.release()
					
					except OSError:
						#TODO: handle this
						print "Error while joining %s with %s." % (a.path, b.path)
						close(fa)
						close(fb)
				
				else:
					break
			
		finally:
			self.joinLock.release()
	
	path = property(getPath)
	size = property(getSize)
	sources = property(getSources)
	onlineSources = property(getOnlineSources)
	status = property(getStatus)
	statusLabel = property(getStatusLabel)
	progress = property(getProgress)

class DCQueueItemSource:
	def __init__(self, nick, path):
		self._nick = nick
		self._path = path

	def getNick(self):
		return self._nick

	def getPath(self, xfer = None):
		return self._path

	nick = property(getNick)
	path = property(getPath)

class DCQueueItemBlock:
	def __init__(self, path, offset, size, completed = False, progress = 0):
		self.path = path
		self.offset = offset
		self.size = size
		
		if completed:
			self.status = DCQueueItem.STATUS_COMPLETED
		else:
			self.status = DCQueueItem.STATUS_WAITING
		
		self.progress = progress
		self.xfer = None

class DCSerializedQueueItem:
	def __init__(self, path, size, sources, status, blocks, progress, blockSerial):
		self.path = path
		self.size = size
		self.sources = sources
		self.status = status
		self.blocks = [ (b.path, b.offset, b.size, b.status == DCQueueItem.STATUS_COMPLETED, b.progress) for b in blocks ]
		self.progress = progress
		self.blockSerial = blockSerial
		
		if self.status == DCQueueItem.STATUS_RUNNING:
			self.status = DCQueueItem.STATUS_WAITING

class IllegalItem(Exception): pass
class IllegalLocalPathError(Exception): pass
class ItemNotMatching(Exception): pass
class ExistingSourceError(Exception): pass
