# This file is part of pyDC software
# Copyright 2002 Anakim Border <aborder@users.sourceforge.net>
#
# pyDC is released under the terms of GPL licence.

from copy import copy
from threading import Lock
import socket

from EventGenerator import *
import DCWorker

class DCSearch(object, EventGenerator):
	LIMIT_NONE = 0
	LIMIT_ATLEAST = 1
	LIMIT_ATMOST = 2

	def __init__(self, private = 0):
		EventGenerator.__init__(self)

		self._pattern = None
		self.patternBlocks = []
		self.sizeLimit = (self.LIMIT_NONE, 0)
		self._private = private
		self._results = []

		self._running = False
		self.timestamp = 0
		self.opLock = Lock()

	def start(self, nodelay = False):
		if self._pattern == None:
			raise PatternNotSetError

		if DCWorker.getWorker().getSearchWorker().registerSearch(self, nodelay):
			self._results = []
			self._running = True

			self.lock.acquire()
			for listener in self.getListeners():
				listener.onSearchStart(self)
			self.lock.release()
		else:
			self._running = False

		return self._running

	def stop(self):
		if self._running:
			DCWorker.getWorker().getSearchWorker().deregisterSearch(self)
			self._running = False

			self.lock.acquire()
			for listener in self.getListeners():
				listener.onSearchStop(self)
			self.lock.release()

	def isAlive(self):
		return self._running

	def onNewResult(self, result):
		try:
			resultPath = result.path.lower()
			for pattern in self.patternBlocks:
				resultPath.index(pattern)

			if self.sizeLimit[0] != self.LIMIT_NONE:
				if self.sizeLimit[0] == self.LIMIT_ATLEAST and result.size < self.sizeLimit[1]:
					raise ValueError
				elif self.sizeLimit[0] == self.LIMIT_ATMOST and result.size > self.sizeLimit[1]:
					raise ValueError
		except ValueError:
			return

		self.opLock.acquire()
		self._results.append(result)
		self.opLock.release()

		self.lock.acquire()
		for listener in self.getListeners():
			listener.onNewResult(self, result)
		self.lock.release()

	# Properties
	def getPattern(self):
		return self._pattern

	def setPattern(self, pattern):
		if self._running: raise SearchInProgressError
		self._pattern = pattern
		self.patternBlocks = pattern.lower().split(' ')
	
	def getSizeLimit(self):
		return self.sizeLimit

	def setSizeLimit(self, tp, size):
		if self._running: raise SearchInProgressError
		if not tp in (self.LIMIT_NONE, self.LIMIT_ATLEAST, self.LIMIT_ATMOST) or \
		  size < 0 or (tp == self.LIMIT_NONE and size != 0):
			raise ValueError

		self.sizeLimit = (tp, size)

	def isPrivate(self):
		return self._private

	def getResultNum(self):
		self.opLock.acquire()
		res = len(self._results)
		self.opLock.release()
		return res

	def getResults(self):
		self.opLock.acquire()
		res = copy(self._results)
		self.opLock.release()
		return res

	#=================#
	# Private section #
	#=================#

	pattern = property(getPattern, setPattern)
	private = property(isPrivate)
	resultNum = property(getResultNum)
	results = property(getResults)

class PatternNotSetError(Exception): pass
class SearchInProgressError(Exception): pass
