# This file is part of pyDC software
# Copyright 2002-2003 Anakim Border <aborder@users.sourceforge.net>
#
# pyDC is released under the terms of GPL licence.

from EventGenerator import *
from Job import *
from socket import socket, error, AF_INET, SOCK_DGRAM
from threading import Lock
from time import time
from Tokenizer import Tokenizer

from DCSearch import *
from DCSearchResult import *

class DCSearchWorker(object, Job, EventGenerator):
	SEARCH_INTERVAL = 10
	SEARCH_TIMEOUT = 300

	def __init__(self, worker):
		EventGenerator.__init__(self)
		self.worker = worker

		self._searches = []
		self.queue = []
		self.timestamp = 0
		self.results = []
		self.sock = socket.socket(AF_INET, SOCK_DGRAM)
		self.opLock = Lock()

	def poll(self):
		self.opLock.acquire()
		t = time()

		# Remove timed out searches
		self._searches = [ s for s in self._searches if t - s.timestamp < self.SEARCH_TIMEOUT ]

		# Handle queued searches
		if len(self.queue) > 0 and t - self.timestamp > self.SEARCH_INTERVAL:
			search = self.queue.pop(0)
			search.timestamp = t
			self._searches.append(search)
			self.timestamp = t

			active = self.worker.getSettings().active
			for hub in self.worker.getHubs():
				hub.startSearch(search, active)

		results = self.results
		self.results = []
		self.opLock.release()

		# Advertise results
		self.lock.acquire()
		for result in results:
			for search in self._searches:
				search.onNewResult(result)
		self.lock.release()

		return False

	def stop(self):
		if self.isAlive():
			self.poll = self.stoppedPoll

	def isAlive(self):
		return self.poll != self.stoppedPoll

	def passiveResult(self, data):
		self.parseResult(data)

	def sendResult(self, addr, data):
		self.sock.sendto(data, addr)
	
	def registerSearch(self, search, nodelay):
		try:
			self.opLock.acquire()

			t = time()
			if (t - self.timestamp < self.SEARCH_INTERVAL or len(self.queue) != 0) and nodelay:
				return False
			else:
				self.queue.append(search)
		finally:
			self.opLock.release()

		self.lock.acquire()
		for listener in self.getListeners():
			listener.onNewSearch(search)
		self.lock.release()

		return True

	def deregisterSearch(self, search):
		self.opLock.acquire()
		try: self._searches.remove(search)
		except:
			try: self.queue.remove(search)
			except: pass
		self.opLock.release()

	def deregisterAllListeners(self):
		self.opLock.acquire()
		for search in self._searches:
			try: search.deregisterAllListeners()
			except: pass
		self.opLock.release()

		EventGenerator.deregisterAllListeners(self)

	def getSearches(self):
		self.opLock.acquire()
		searches = copy(self._searches)
		self.opLock.release()
		return searches

	#=================#
	# Private section #
	#=================#

	def stoppedPoll(self):
		return True

	def parseResult(self, data):
		#TODO: correctly parse directory results
		try:
			t = Tokenizer(data)
			r = DCSearchResult()

			r.userNick = t.next(' ')
			r.path = t.next('\5')
			r.type = r.path[r.path.rindex('.')+1:]
			r.size = long(t.next(' '))
			r.freeSlots = int(t.next('/'))
			r.slots = int(t.next('\5'))

			r.hubName = t.flush()
			r.hubName = r.hubName[:r.hubName.rindex('(')-1]

			self.opLock.acquire()
			self.results.append(r)
			self.opLock.release()

		except ValueError:
			return None

	searches = property(getSearches, None, None, None)
