# This file is part of pyDC software
# Copyright 2002-2003 Anakim Border <aborder@users.sourceforge.net>
#
# pyDC is released under the terms of GPL licence.

from copy import copy
from os import access, F_OK, R_OK, W_OK, linesep
from os.path import basename, isdir
from socket import inet_aton, error
from types import ClassType
from urlparse import urlparse
from xml.sax.saxutils import escape, unescape
from weakref import proxy
import xml.sax
import xml.sax.handler

from DCUser import DCUser

class DCSettings(object):
	DEFAULT_HUBLIST = 'http://www.neo-modus.com/PublicHubList.config'
	DEFAULT_PORT = 415
	DEFAULT_SLOTNUM = 3

	def __init__(self):
		self.reset()
		self._error = 1 #invalid configuration: default settings are not enough
		
		# Default values
		self._hubLists = [self.DEFAULT_HUBLIST]
		self._port = self.DEFAULT_PORT
		self._slots = self.DEFAULT_SLOTNUM
	
	def load(self, path):
		self.reset()
		
		try:
			parser = xml.sax.make_parser()
			parser.setContentHandler(MainHandler(proxy(self), parser))
			parser.parse(path)
		except IOError:
			raise ParseError(ParseError.E_IO)
		except xml.sax.SAXException:
			raise ParseError(ParseError.E_XML)
		
		if self._error: raise self._error

	def save(self, path):
		o = file(path, 'w')

		self.writeTag(o, 'settings', 0)
		self.writeBlock(o, 'nick', self.user.nick, 1)
		self.writeBlock(o, 'connection', DCUser.CONN_NAMES[self.user.connection], 1)
		self.writeBlock(o, 'description', self.user.description, 1)
		self.writeBlock(o, 'email', self.user.email, 1)

		self.writeBlock(o, 'download', self._downloadDir, 1)
		if len(self._sharedDirs) == 0:
			self.writeTag(o, 'share/', 1)
		else:
			self.writeTag(o, 'share', 1)
			for d in self._sharedDirs:
				self.writeBlock(o, 'dir', d, 2)
			self.writeTag(o, '/share', 1)

		if len(self._hubLists) == 0:
			self.writeTag(o, 'hublist/', 1)
		else:
			self.writeTag(o, 'hublist', 1)
			for source in self._hubLists:
				self.writeBlock(o, 'source', source, 2)
			self.writeTag(o, '/hublist', 1)
		
		if len(self._favorites) == 0:
			self.writeTag(o, 'favorites/', 1)
		else:
			self.writeTag(o, 'favorites', 1)
			for fav in self._favorites:
				self.writeTag(o, 'favorite', 2)
				self.writeBlock(o, 'address', fav[0], 3)
				self.writeBlock(o, 'autoconnect', str(fav[1]), 3)
				self.writeBlock(o, 'nick', fav[2], 3)
				self.writeBlock(o, 'password', fav[3], 3)
				self.writeTag(o, '/favorite', 2)
			self.writeTag(o, '/favorites', 1)

		self.writeBlock(o, 'ip', self._ip, 1)
		self.writeBlock(o, 'port', str(self._port), 1)
		
		if self._active:
			self.writeBlock(o, 'mode', 'active', 1)
		else:
			self.writeBlock(o, 'mode', 'passive', 1)

		self.writeBlock(o, 'slots', str(self._slots), 1)
		self.writeBlock(o, 'uploadBandwidth', str(self.getUploadBandwidth()), 1)
		self.writeBlock(o, 'downloadBandwidth', str(self.getDownloadBandwidth()), 1)
		
		if len(self._custom):
			self.writeTag(o, 'custom', 1)
			for key, value in self._custom.iteritems():
				self.writeBlock(o, key, str(value), 2)
			self.writeTag(o, '/custom', 1)
		else:
			self.writeTag(o, 'custom/', 1)

		self.writeTag(o, '/settings', 0)
		o.close()
		
	def setError(self, error):
		if self._error == None:
			self._error = error
	
	def isValid(self):
		return self._error == None
	
	def getNick(self):
		return self.user.nick

	def setNick(self, nick):
		if not self.validateNick(nick):
			raise ValueError

		self.user.nick = nick

	def getConnection(self):
		return self.user.connection

	def setConnection(self, connection):
		if not self.validateConnection(connection):
			raise ValueError

		self.user.connection = connection

	def getDescription(self):
		return self.user.description

	def setDescription(self, description):
		if not self.validateDescription(description):
			raise ValueError

		self.user.description = description

	def getEmail(self):
		return self.user.email
		
	def setEmail(self, email):
		if not self.validateEmail(email):
			raise ValueError

		self.user.email = email

	def getDownloadDir(self):
		return self._downloadDir
	
	def setDownloadDir(self, dir):
		if not self.validateDownloadDir(dir):
			raise ValueError

		self._downloadDir = dir

	def getSharedDirs(self):
		return copy(self._sharedDirs)
	
	def setSharedDirs(self, list):
		self._sharedDirs = []
		
		for s in list:
			if not self.validateSharedDir(s, self._sharedDirs): raise ValueError
			self.addSharedDir(s)
	
	def clearSharedDirs(self):
		self._sharedDirs = []
	
	def addSharedDir(self, dir):
		if not self.validateSharedDir(dir, self._sharedDirs):
			raise ValueError
		
		self._sharedDirs.append(dir)

	def getHubLists(self):
		return copy(self._hubLists)
		
	def setHubLists(self, lists):
		self._hubLists = []
		
		for l in lists:
			if not self.validateHubList(l): raise ValueError
			self.addHubList(l)
	
	def clearHubLists(self):
		self._hubLists = []
	
	def addHubList(self, list):
		if not self.validateHubList(list):
			raise ValueError
		
		self._hubLists.append(list)
	
	def getFavorites(self):
		return copy(self._favorites)
	
	def getFavorite(self, addr):
		for f in self._favorites:
			if f[0] == addr:
				return copy(f)
		return None
		
	def clearFavorites(self):
		self._favorites = []
	
	def addFavorite(self, address, autoconnect, nick, password):
		if nick and len(nick) == 0: nick = None
		if password and len(password) == 0: password = None
	
		if not self.validateFavorite(address, nick, password):
			raise ValueError
		
		self._favorites.append((address, autoconnect, nick, password))
	
	def updateFavorite(self, oldaddr, addr, autoconnect, nick, password):
		if nick and len(nick) == 0: nick = None
		if password and len(password) == 0: password = None
	
		if not self.validateFavorite(addr, nick, password):
			raise ValueError
		
		aux = []
		for f in self._favorites:
			if f[0] == oldaddr:
				aux.append((addr, autoconnect, nick, password))
			else:
				aux.append(f)
		
		self._favorites = aux
		
	def removeFavorite(self, address):
		self._favorites = [ f for f in self._favorites if f[0] != address ]

	def getIP(self):
		return self._ip
	
	def setIP(self, ip):
		if not self.validateIP(ip):
			raise ValueError
	
		self._ip = ip

	def getPort(self):
		return self._port
	
	def setPort(self, port):
		if not self.validatePort(port):
			raise ValueError

		self._port = port

	def isActive(self):
		return self._active

	def setActive(self, active):
		self._active = active

	def getSlotNum(self):
		return self._slots
	
	def setSlotNum(self, num):
		if not self.validateSlotNum(num):
			raise ValueError
			
		self._slots = num
	
	def getUploadBandwidth(self):
		return self._uploadBandwidth
	
	def setUploadBandwidth(self, bps):
		if not self.validateUploadBandwidth(bps):
			raise ValueError
		
		self._uploadBandwidth = bps
	
	def getDownloadBandwidth(self):
		return self._downloadBandwidth
	
	def setDownloadBandwidth(self, bps):
		if not self.validateDownloadBandwidth(bps):
			raise ValueError
		
		self._downloadBandwidth = bps
	
	def getCustom(self, key, default = None):
		return self._custom[key]
		
	def getCustomEx(self, key, type = str, default = None):
		if self._custom.has_key(key):
			try:
				value = self._custom[key]
				value = type(value)
			except:
				value = default
			
			self._custom[key] = value
			return value
		
		else:
			self._custom[key] = default
			return default
	
	def setCustom(self, key, value):
		self._custom[key] = value
	
	def deleteCustom(self, key):
		if self._custom.has_key(key):
			del self._custom[key]
	
	# Validators
	def validateNick(nick):
		return len(nick) <= 32 and DCSettings.checkChars(nick)
	
	def validateConnection(connection):
		return DCUser.CONN_NAMES.has_key(connection) and connection != DCUser.CONN_NONE
	
	def validateDescription(description):
		return DCSettings.checkChars(description)
	
	def validateEmail(email):
		return DCSettings.checkChars(email) and email.find('@') != -1
	
	def validateDownloadDir(dir):
		return isdir(dir) and access(dir, F_OK | R_OK | W_OK)
	
	def validateSharedDir(dir, dirList = []):
		if not isdir(dir) or not access(dir, F_OK | R_OK):
			return False
		
		base = basename(dir)
		for d in dirList:
			if base == basename(d):
				return False
		
		return True
	
	def validateHubList(list):
		url = urlparse(list)
		return url[0] == 'http' and len(url[1]) > 0
	
	def validateFavorite(address, nick, password):
		if address == None:
			return False
		
		sep = address.rfind(':')
		if sep > 0:
			try:
				port = int(address[sep+1:])
				if port < 0 or port > 0xffff:
					return False
			except ValueError:
				return False
		elif sep == 0:
			return False
			
		return nick == None or len(nick) == 0 or DCSettings.validateNick(nick)
	
	def validateIP(ip):
		if ip == None: return True
		
		try:
			inet_aton(ip)
			return True
		except error:
			return False
	
	def validatePort(port):
		return port > 0 and port < 0xffff
	
	def validateSlotNum(num):
		return num > 0 and num < 100
	
	def validateUploadBandwidth(bps):
		return bps >= 0
	
	def validateDownloadBandwidth(bps):
		return bps >= 0
	
	###################
	# Private members #
	###################

	def reset(self):
		self._error = None
	
		# No default values here
		self.user = DCUser()
		self._downloadDir = None
		self._sharedDirs = []
		self._hubLists = []
		self._favorites = []

		self._ip = None
		self._port = 0
		
		self._active = True
		self._slots = -1
		self._uploadBandwidth = 0.0
		self._downloadBandwidth = 0.0
		
		self._custom = {}

	def checkChars(content):
		if content.find('|') != -1 or content.find('$') != -1:
			return False
		return True

	def writeTag(self, o, tag, level):
		data  = '\t' * level
		data += '<%s>' % tag
		data += linesep
		o.write(data)

	def writeBlock(self, o, tag, value, level):
		data  = '\t' * level
		if value != None and len(value) > 0:
			data += '<%s>%s</%s>' % (tag, escape(value), tag)
		else:
			data += '<%s/>' % tag
		data += linesep
		o.write(data)
	
	# Static methods
	validateNick = staticmethod(validateNick)
	validateConnection = staticmethod(validateConnection)
	validateDescription = staticmethod(validateDescription)
	validateEmail = staticmethod(validateEmail)
	validateDownloadDir = staticmethod(validateDownloadDir)
	validateSharedDir = staticmethod(validateSharedDir)
	validateHubList = staticmethod(validateHubList)
	validateFavorite = staticmethod(validateFavorite)
	validateIP = staticmethod(validateIP)
	validatePort = staticmethod(validatePort)
	validateSlotNum = staticmethod(validateSlotNum)
	validateUploadBandwidth = staticmethod(validateUploadBandwidth)
	validateDownloadBandwidth = staticmethod(validateDownloadBandwidth)
	checkChars = staticmethod(checkChars)
	
	# Properties
	valid = property(isValid)
	nick = property(getNick, setNick)
	connection = property(getConnection, setConnection)
	description = property(getDescription, setDescription)
	email = property(getEmail, setEmail)
	downloadDir = property(getDownloadDir, setDownloadDir)
	sharedDirs = property(getSharedDirs, setSharedDirs)
	hubLists = property(getHubLists, setHubLists)
	favorites = property(getFavorites)
	ip = property(getIP, setIP)
	port = property(getPort, setPort)
	active = property(isActive, setActive)
	slotNum = property(getSlotNum, setSlotNum)
	uploadBandwidth = property(getUploadBandwidth, setUploadBandwidth)
	downloadBandwidth = property(getDownloadBandwidth, setDownloadBandwidth)

class MainHandler(xml.sax.handler.ContentHandler):
	def __init__(self, repository, parser):
		self.repository = repository
		self.parser = parser

	def endDocument(self):
		# Do we have a minimal configuration?
		if self.repository.user.nick == None:
			raise ParseError(ParseError.E_NONICK)
		elif self.repository.user.connection == DCUser.CONN_NONE:
			raise ParseError(ParseError.E_NOCONNECTION)
		elif self.repository._downloadDir == None:
			raise ParseError(ParseError.E_NODOWNLOAD)
		elif self.repository._active and self.repository._port == 0:
			raise ParseError(ParseError.E_NOPORT)
		elif self.repository._slots == -1:
			raise ParseError(ParseError.E_NOSLOTNUM)
		
	def startElement(self, name, attrs):
		# The only element we expect here is 'settings'
		if name != 'settings':
			raise ParserError(ParseError.E_SYNTAX)

		# Setup 'settings' content handler
		handler = SettingsHandler(self.repository, self.parser, self)
		self.parser.setContentHandler(handler)

class BaseHandler(xml.sax.handler.ContentHandler):
	def __init__(self, repository, parser, parent):
		self.repository = repository
		self.parser = parser
		self.parent = parent
		self.value = None

	def characters(self, content):
		if self.value == None: return

		content = content.encode('ascii')
		if len(self.value) == 0:
			self.value = content
		else:
			if self.value[-1] in (' ', '\t'):
				content = content.lstrip()
			self.value += content

class SettingsHandler(BaseHandler):
	def __init__(self, repository, parser, parent):
		BaseHandler.__init__(self, repository, parser, parent)
		self.elemInfo = None
		
		self.elements = { 'nick' :               (repository.setNick, None, ParseError.E_NICK),
		                  'connection' :         (repository.setConnection, self.connectionFilter, ParseError.E_CONNECTION),
		                  'description' :       (repository.setDescription, None, ParseError.E_DESCRIPTION),
		                  'email' :             (repository.setEmail, None, ParseError.E_EMAIL),
		                  'download' :          (repository.setDownloadDir, None, ParseError.E_DOWNLOAD),
		                  'ip' :                (repository.setIP, None, ParseError.E_IP),
		                  'port' :              (repository.setPort, int, ParseError.E_PORT),
		                  'mode' :              (repository.setActive, self.modeFilter, ParseError.E_MODE),
		                  'slots' :             (repository.setSlotNum, int, ParseError.E_SLOTNUM),
		                  'uploadBandwidth' :   (repository.setUploadBandwidth, float, ParseError.E_UPDBAND),
		                  'downloadBandwidth' : (repository.setDownloadBandwidth, float, ParseError.E_DOWNBAND)
	                    }
		
		self.parsers = { 'share' : ShareHandler,
		                 'hublist' : HubListHandler,
		                 'favorites' : FavoritesHandler,
		                 'custom' : CustomHandler }

	def startElement(self, name, attrs):
		try:
			self.elemInfo = self.elements[name]
			self.value = ''
		except KeyError:
			if self.parsers.has_key(name):
				self.parser.setContentHandler(self.parsers[name](self.repository, self.parser, self))
			else:
				raise ParseError(ParseError.E_SYNTAX)
	
	def endElement(self, name):
		if name == 'settings':
			# This block has ended; return control to the parent handler
			return self.parser.setContentHandler(self.parent)

		value = self.value.strip()
		if len(value) > 0:
			try:
				if self.elemInfo[1] != None:
					value = self.elemInfo[1](value)

				self.elemInfo[0](value)
			except ValueError:
				self.repository.setError(ParseError(self.elemInfo[2]))

		self.elemInfo = None
		self.value = None
		
	def connectionFilter(self, value):
		for connection, name in DCUser.CONN_NAMES.iteritems():
			if name == value:
				return connection
		raise ValueError

	def modeFilter(self, value):
		value = value.lower().strip()
		if value == 'active':
			return True
		elif value == 'passive':
			return False
		else:
			raise ValueError
			
class ShareHandler(BaseHandler):
	def __init__(self, repository, parser, parent):
		BaseHandler.__init__(self, repository, parser, parent)

	def startElement(self, name, attrs):
		if name != 'dir':
			raise ParseError(ParseError.E_SYNTAX)
			
		self.value = ''
		
	def endElement(self, name):
		if name == 'share':
			# This block has ended; return control to the parent handler
			return self.parser.setContentHandler(self.parent)

		value = self.value.strip()
		self.value = None
		if len(value) == 0:
			raise ParseError(ParseError.E_SYNTAX)

		try:
			self.repository.addSharedDir(value)
		except ValueError:
			self.repository.setError(ParseError(ParseError.E_SHARE))

class HubListHandler(BaseHandler):
	def __init__(self, repository, parser, parent):
		BaseHandler.__init__(self, repository, parser, parent)

	def startElement(self, name, attrs):
		if name != 'source':
			raise ParseError(ParseError.E_SYNTAX)

		self.value = ''
		
	def endElement(self, name):
		if name == 'hublist':
			# This block has ended; return control to the parent handler
			return self.parser.setContentHandler(self.parent)

		value = self.value.strip()
		self.value = None
		if len(value) == 0:
			self.repository.setError(ParseError(ParseError.E_SYNTAX))
		else:
			try:
				self.repository.addHubList(value)
			except ValueError:
				self.repository.setError(ParseError(ParseError.E_HUBLIST))
		
class FavoritesHandler(BaseHandler):
	def __init__(self, repository, parser, parent):
		BaseHandler.__init__(self, repository, parser, parent)

	def startElement(self, name, attrs):
		if not name in 'favorite':
			raise ParseError(ParseError.E_SYNTAX)
		
		self.parser.setContentHandler(FavoriteHandler(self.repository, self.parser, self))
		
	def endElement(self, name):
		if name == 'favorites':
			# This block has ended; return control to the parent handler
			return self.parser.setContentHandler(self.parent)

class FavoriteHandler(BaseHandler):
	def __init__(self, repository, parser, parent):
		BaseHandler.__init__(self, repository, parser, parent)
		self.tag = None
		self.address = None
		self.autoconnect = True
		self.nick = None
		self.password = None

	def startElement(self, name, attrs):
		if not name in ('address', 'autoconnect', 'nick', 'password'):
			raise ParseError(ParseError.E_SYNTAX)
		
		self.tag = name
		self.value = ''
		
	def endElement(self, name):
		if name == 'favorite':
			try:
				self.repository.addFavorite(self.address, self.autoconnect, self.nick, self.password)
			except ValueError:
				self.repository.setError(ParseError(ParseError.E_FAVORITE))
			
			# This block has ended; return control to the parent handler
			return self.parser.setContentHandler(self.parent)

		value = self.value.strip()
		self.value = None
		if self.tag in ('address', 'autoconnect') and len(value) == 0:
			self.repository.setError(ParseError(ParseError.E_FAVORITE))
		else:
			if self.tag == 'address':
				self.address = value
				
			elif self.tag == 'autoconnect':
				value = value.lower()
				if value in ('1', 'true', 'yes'):
					self.autoconnect = True
				elif value in ('0', 'false', 'no'):
					self.autoconnect = False
				else:
					self.repository.setError(ParseError(ParseError.E_FAVCONNECT))
				
			elif self.tag == 'nick':
				self.nick = value
				
			elif self.tag == 'password':
				self.password = value

class CustomHandler(BaseHandler):
	def __init__(self, repository, parser, parent):
		BaseHandler.__init__(self, repository, parser, parent)
		self.key = None
		self.value = None

	def startElement(self, name, attrs):
		self.key = name
		self.value = ''
		
	def endElement(self, name):
		if name == 'custom':
			# This block has ended; return control to the parent handler
			return self.parser.setContentHandler(self.parent)

		value = self.value.strip()
		self.value = None
		
		self.repository._custom[self.key] = value


class ParseError:
	E_IO             = 1
	E_XML            = 2
	E_SYNTAX         = 3
	E_NICK           = 4
	E_CONNECTION     = 5
	E_DESCRIPTION    = 6
	E_EMAIL          = 7
	E_DOWNLOAD       = 8
	E_SHARE          = 9
	E_HUBLIST        = 10
	E_MODE           = 11
	E_IP             = 12
	E_PORT           = 13
	E_SLOTNUM        = 14
	E_UPDBAND        = 15
	E_DOWNBAND       = 16
	E_NONICK         = 17
	E_NOCONNECTION   = 18
	E_NODOWNLOAD     = 19
	E_NOPORT         = 20
	E_NOSLOTNUM      = 21
	E_FAVORITE       = 22
	E_FAVCONNECT     = 23

	descriptions = { E_IO :           'read failure',
	                 E_XML :          'malformed XML',
	                 E_SYNTAX :       'invalid syntax',
	                 E_NICK :         'invalid nick',
	                 E_CONNECTION :   'invalid connection',
	                 E_DESCRIPTION :  'invalid description',
	                 E_EMAIL :        'invalid email address',
	                 E_DOWNLOAD :     'invalid download directory (chek you can write inside it)',
	                 E_SHARE :        'invalid share directory',
	                 E_HUBLIST :      'invalid hub list source',
	                 E_MODE :         'invalid work mode',
	                 E_IP :           'invalid IP address',
	                 E_PORT :         'invalid port',
	                 E_SLOTNUM :      'invalid slot number',
	                 E_UPDBAND :      'invalid upload bandwidth',
	                 E_DOWNBAND :     'invalid download bandwidth',
	                 E_NONICK :       'no nick found',
	                 E_NOCONNECTION : 'no connection type found',
	                 E_NODOWNLOAD :   'no download location found',
	                 E_NOPORT :       'no port number found',
	                 E_NOSLOTNUM :    'no slot number found',
	                 E_FAVORITE :     'invalid favorite hub',
	                 E_FAVCONNECT :   'invalid autoconnect value' }

	def __init__(self, code):
		self.code = code

	def __repr__(self):
		return "Error: " + ParseError.descriptions[self.code] + "."

	def getDescription(self):
		return ParseError.descriptions[self.code]

if __name__ == '__main__':
 	s = DCSettings()
 	try:
 		s.load('../settings.xml')
 	except ParseError, e:
 		print "==> ParseError"
 		print "   ", e.getDescription()

