# This file is part of pyDC software
# Copyright 2002-2004 Anakim Border <aborder@users.sourceforge.net>
#
# pyDC is released under the terms of GPL licence.

from os import access, R_OK, error, sep
from os.path import basename, exists, getsize, isfile
from time import time
from Tokenizer import Tokenizer
from zlib import compressobj

from Channel import ChannelHandler
from DCChannel import DCChannelHandler
from DCItem import DCItem
import DCWorker

class DCUploadSetupHandler(DCChannelHandler):
	HANDSHAKE_TIMEOUT = 30

	def __init__(self):
		cmds = { 'GetListLen' :    ( 0, DCUploadSetupHandler.onGetListLen),
		         'Get' :           (-1, DCUploadSetupHandler.onGet),
		         'Send' :          ( 0, DCUploadSetupHandler.onSend),
		         'GetTestZBlock' : (-1, DCUploadSetupHandler.onGetZBlock),
		         'GetZBlock' :     (-1, DCUploadSetupHandler.onGetZBlock) }
		DCChannelHandler.__init__(self, cmds)
	
	def onAttach(self, channel):
		DCChannelHandler.onAttach(self, channel)
		self.timestamp = time()
	
	def poll(self):
		if time() - self.timestamp > self.HANDSHAKE_TIMEOUT:
			raise ValueError, '[upload] handshake timeout'
		
		try:
			DCChannelHandler.poll(self)
		except Exception, e:
			self.channel.sendCmd('Error')
			raise e

	def onInvalidCommand(self, cmd):
		raise ValueError

	def onUnknownCommand(self, cmd):
		self.channel.sendCmd('Error')

	def onChatMessage(self, cmd):
		raise ValueError

	def onGetListLen(self):
		self.channel.sendCmd('ListLen', str(getsize(DCWorker.getWorker().FILELIST_HE3)))

	def onGet(self, data):
		name, pos = data.split('$')
		pos = long(pos)-1
		if pos < 0:
			return ValueError
		
		item = self.findItem(name)
		if item == None:
			return self.channel.sendCmd('Error')
		
		if DCWorker.getWorker().acquireSlot():
			self.channel.registerCleanup(self.cleanup, None)
		else:
			return self.channel.sendCmd('MaxedOut')
		
		self.channel.file = file(item.path, 'rb')
		self.channel.file.seek(pos, 0)
		self.channel.sendCmd('FileLength', str(item.size))

		self.channel._item = item
		self.channel._sizeRemaining = item.size - self.channel.file.tell()
	
	def onSend(self):
		self.channel.setHandler(DCUploadTransferHandler())
	
	def onGetZBlock(self, cmd):
		t = Tokenizer(cmd)
		offset = t.next()
		size = t.next()
		name = t.flush()
		
		offset = long(offset)
		size = long(size)
		if offset < 0 or size < 0:
			return ValueError
		
		item = self.findItem(name)
		if item == None:
			return self.channel.sendCmd('Failed', 'No such file')
			
		if offset > item.size:
			print "Bad offset '%d' for file '%s'" % (offset, name)
			return self.channel.sendCmd('Failed', 'Block offset too big')
			
		if offset + size > item.size:
			print "Bad size '%d' for file '%s'" % (size, name)
			return self.channel.sendCmd('Failed', 'Block size too big')
		
		item.size = size
		
		if DCWorker.getWorker().acquireSlot():
			self.channel.registerCleanup(self.cleanup, None)
		else:
			return self.channel.sendCmd('MaxedOut')
		
		self.channel.file = file(item.path, 'rb')
		self.channel.file.seek(offset, 0)

		self.channel._item = item
		self.channel._sizeRemaining = size
		self.channel._compressed = True
		
		self.channel.sendCmd('Sending')
		self.channel.setHandler(DCUploadTransferHandler())
		
	def findItem(self, name):
		item = DCItem()
		worker = DCWorker.getWorker()
		settings = worker.getSettings()
		
		if name in (worker.FILELIST_HE3, worker.FILELIST_BZ2):
			item.path = name
			item.size = getsize(name)
			f = file(name, 'rb')
		
		else:
			idx = name.index('\\')
			root, path = name[:idx], name[idx+1:].replace('\\', sep)
			f = False
			
			for dir in settings.getSharedDirs():
				if basename(dir) != root:
					continue
			
				item.path = dir + sep + path
				if isfile(item.path):
					item.size = getsize(item.path)
					f = True
					break
			
			if not f:
				return None
		
		item.userNick = settings.getNick()
		return item
	
	def cleanup(dummy):
		DCWorker.getWorker().releaseSlot()
	
	cleanup = staticmethod(cleanup)

class DCUploadTransferHandler(ChannelHandler):
	SPEEDSTAT_TIMEOUT = 3
	
	def __init__(self):
		ChannelHandler.__init__(self)
		
	def onAttach(self, xfer):
		ChannelHandler.onAttach(self, xfer)
		self.speed = self.channel._speed
		self.sock = self.channel.sock
		
		if self.channel._compressed:
			self.zlib = ZlibReader(self.channel.file, self.channel._item.size)
		else:
			self.zlib = None
		
		self.channel.notify()
		self.timestamp = self.statstamp = time()
		self.speed.init(self.timestamp)
	
	def prepoll(self):
		self.timestamp = time()
		bytes = self.speed.getBandwidth(self.timestamp)
		queuedBytes = self.sock.getBytesToSend()
		
		if queuedBytes < bytes:
			delta = long(bytes - queuedBytes)
			if self.zlib == None:
				self.sock.send(self.channel.file.read(min(delta<<4, self.channel._sizeRemaining)))
			else:
				self.sock.send(self.zlib.read(delta))
	
	def poll(self):
		delta = time()
	
		netLen = self.sock.getBytesSent()
		if self.zlib:
			dataLen = self.zlib.consumed()
			ratio = self.zlib.ratio()
		else:
			dataLen = netLen
			ratio = 1
		
		self.speed.usedBandwidth(self.timestamp, netLen, netLen*ratio)
		self.channel._sizeRemaining -= dataLen

		if self.timestamp - self.statstamp > self.SPEEDSTAT_TIMEOUT:
			self.channel._progress = (1 - float(self.channel._sizeRemaining) / self.channel._item.size) * 100
			self.statstamp = self.timestamp
			DCWorker.getWorker().getXferWorker().onXferUpdate(self.channel)
		
		if self.channel._sizeRemaining <= 0 and self.sock.getBytesToSend() == 0:
			if self.zlib and not self.zlib.eof():
				return
			
			if len(self.channel.cleanup):
				DCWorker.getWorker().releaseSlot()
			self.channel.nextItem()

class ZlibReader:
	def __init__(self, source, size):
		self.compressor = compressobj()
		self.source = source
		self.remainingBytes = size
		self._consumed = 0
		self._ratio = 1
		self.flushed = False
		self.buf = ''
	
	def read(self, size):
		self._consumed = 0
		compressed = 0
		
		while True:
			if len(self.buf) >= size:
				break
			elif self.remainingBytes == 0:
				if not self.flushed:
					self.buf += self.compressor.flush()
					self.flushed = True
				break
			
			data = self.source.read(min(size, self.remainingBytes))
			dlen = len(data)
			if dlen == 0:
				#DEBUG
				print "WARNING: found EOF while remainingBytes =", self.remainingBytes
				break
			
			self._consumed += dlen
			self.remainingBytes -= dlen
			
			zdata = self.compressor.compress(data)
			self.buf += zdata
			compressed += len(zdata)
			
		data = self.buf[:size]
		self.buf = self.buf[size:]
		
		if self._consumed:
			self._ratio = float(self._consumed) / compressed
		
		return data
		
	def consumed(self):
		consumed = self._consumed
		self._consumed = 0
		return consumed
	
	def ratio(self):
		return self._ratio
	
	def eof(self):
		return self.remainingBytes == 0 and len(self.buf) == 0
