# This file is part of pyDC software
# Copyright 2002-2004 Anakim Border <aborder@users.sourceforge.net>
#
# pyDC is released under the terms of GPL licence.

from bz2 import BZ2File
from EventGenerator import *
from he3 import *
from Job import *
from os import sep, unlink
from thread import start_new_thread

from DCItem import *
from DCFileList import *
from DCQueueItem import *
from DCQueueEventListener import *
import DCWorker

class DCUserList(DCQueueItem, DCQueueEventListener):
	FILENAME_TEMPLATE = 'tmp%d.lst'
	_numLock = Lock()
	_tmpnum = 0
	
	def __init__(self, user):
		if not user.hub:
			raise UserZombieError
		
		DCUserList._numLock.acquire()
		filename = DCUserList.FILENAME_TEMPLATE % DCUserList._tmpnum
		DCUserList._tmpnum += 1
		DCUserList._numLock.release()

		self._list = []
		self._user = user
		self.fullPath = DCWorker.getWorker().getSettings().downloadDir + sep + filename
		self.decoding = False
		
		try: unlink(self.fullPath)
		except OSError: pass
		
		item = DCItem()
		item.userNick = user.nick
		item.path = DCWorker.getWorker().FILELIST_HE3
		
		DCQueueItem.__init__(self, item, filename)
		self.getSources()[0].__class__ = DCUserListSource
		
		DCWorker.getWorker().getQueue().addQueueItem(self, True)
		DCWorker.getWorker().getQueue().registerListener(self)
		
	def addSource(self, item):
		pass
	
	def removeSource(self, nick, stop = True, cause = 0):
		self.lock.acquire()
		signal, dequeue = False, False
		if nick == self._user.nick and self._status != self.STATUS_ERROR:
			if cause:
				self._status = self.STATUS_ERROR
				signal = True
			else:
				dequeue = True
		self.lock.release()
		
		worker = DCWorker.getWorker()
		if stop: worker.getXferWorker().stopXfer(self)
		if signal: worker.getQueue().onItemStatus(self)
		if dequeue: worker.getQueue().removeItem(self)
	
	def serialize(self):
		return None
	
	def needsDownload(self):
		return self._status == self.STATUS_WAITING and DCQueueItem.needsDownload(self)
	
	def getBlock(self, xfer):
		return DCQueueItem.getBlock(self, xfer, True)
	
	def getList(self):
		return self._list
	
	def getUser(self):
		return self._user

	def getPattern(self):
		return None
	
	def getStatusLabel(self):
		if self._status == self.STATUS_RUNNING and self.decoding:
			return 'Decoding'
		else:
			return DCQueueItem.getStatusLabel(self)

 	def getProgress(self):
		if self._status != self.STATUS_COMPLETED:
 			return DCQueueItem.getProgress(self) * 95 / 100
		else:
			return 100.0
	
	def setStatusCompleted(self, source, block):
		w = DCWorker.getWorker()
		
		self.decoding = True
		w.getQueue().onItemStatus(self)
		bz2 = self.getSources()[0].getPath() == w.FILELIST_BZ2
		start_new_thread(self.parseList, (bz2,))

	def sendRequest(self, lockedSources):
		if not self._user.hub:
			#TODO: do we need the following?
# 			self.setStatus(self.STATUS_ERROR)
			DCWorker.getWorker().getQueue().removeItem(self)
			return
		
		self._user.hub.requireConnection(self._user.nick)
	
	def sort(self):
		self._list = self.__sort(self._list)
	
	###################
	# Private methods #
	###################

	def parseList(self, bz2):
		try:
			if bz2:
				data = BZ2File(self.fullPath, 'r').read()
			else:
				data = He3Decoder().decode(self.fullPath)
			
			r = LineReader(data)
			self._list = self.populate(r, 0, self)
			self._status = self.STATUS_COMPLETED
		except Exception:
			self._status = self.STATUS_ERROR
		
		DCWorker.getWorker().getQueue().onItemStatus(self)
		
		try: unlink(self.fullPath)
		except OSError: pass

	def populate(self, r, level, parent):
		l = []

		while True:
			line = r.get()
			if line == None:
				return l

			lineLevel = 0
			while line[lineLevel] == '\t':
				lineLevel += 1

			if lineLevel < level:
				r.unget()
				return l
			else:
				pos = line.rfind('|')
				if pos == -1:
					d = UserDir(line[lineLevel:], parent, None)
					d.children = self.populate(r, level+1, d)
					l.append(d)
				else:
					l.append(UserFile(line[lineLevel:pos], parent, long(line[pos+1:])))
	
	def __sort(self, list):
		for pos, item in enumerate(list):
			if isinstance(item, UserDir):
				list[pos].children = self.__sort(item.children)
		
		aux = [ (i.name, i) for i in list ]
		aux.sort()
		return [ a[1] for a in aux ]
	
	list = property(getList)
	user = property(getUser)

class DCUserListSource(DCQueueItemSource):
	def getPath(self, xfer = None):
		if xfer and xfer.supports('BZList'):
			self._path = DCWorker.getWorker().FILELIST_BZ2
			return self._path
		else:
			return self._path

class LineReader:
	def __init__(self, data):
		self.lines = data.split('\r\n')
		if len(self.lines[-1]) == 0:
			self.lines = self.lines[:-1]
		self.pos = 0

	def get(self):
		if self.pos >= len(self.lines):
			return None

		line = self.lines[self.pos]
		self.pos += 1
		return line

	def unget(self):
		self.pos -= 1

class UserDir:
	def __init__(self, name, parent, children):
		self.name = name
		self.children = children
		
		if parent == None:
			self.parent = None
		else:
			self.parent = proxy(parent)

	def getChildren(self):
		return self.children
	
	c = property(getChildren)

class UserFile:
	def __init__(self, name, parent, size):
		self.name = name
		self.parent = proxy(parent)
		self.size = size
		
	def getPath(self):
		path, item = self.name, self.parent
		while isinstance(item, UserDir):
			path = item.name + '\\' + path
			item = item.parent
		return path
		
	def getUserNick(self):
		item = self.parent
		while isinstance(item, UserDir): item = item.parent
		return item.nick

	def fullInfo(self):
		path, item = self.name, self.parent
		while isinstance(item, UserDir):
			path = item.name + '\\' + path
			item = item.parent
		return item.getUser().nick, path, self.size

	def __str__(self):
		res  = "=========\n"
		res += "Name:" + self.name + "\n"
		res += "Size: " + str(self.size)
		return res

	path = property(getPath)
	userNick = property(getUserNick)

class UserZombieError(Exception): pass
