# This file is part of pyDC software
# Copyright 2002-2004 Anakim Border <aborder@users.sourceforge.net>
#
# pyDC is released under the terms of GPL licence.

from copy import copy
from EventGenerator import *
from JobPools import *
from thread import start_new_thread
from threading import Lock
from time import time, sleep
import atexit
import socket
import sys

from DCHub import *
from DCLocalIP import *
from DCLocalList import *
from DCLocalListWriter import *
from DCQueue import *
from DCSearchListener import *
from DCSearchWorker import *
from DCSettings import *
from DCUser import DCUser
from DCXferListener import *
from DCXferWorker import *
from DCWorkerEventListener import *

class DCWorker(object, EventGenerator):
	PYDCLIB_ID           = 'pyDC'
	PYDCLIB_VERSION      = '0.7'
	PYDCLIB_DCPP_VERSION = '0.304'

	STATUS_READY = 0
	STATUS_RUNNING = 1
	STATUS_STOPPING = -1
	STATUS_STOPPED = -2
	
	XFER_GRACE = 45
	
	FILELIST_HE3 = 'MyList.DcList'
	FILELIST_BZ2 = 'MyList.bz2'

	def __init__(self):
		EventGenerator.__init__(self)
	
 		self.opLock          = Lock()
		self.status          = self.STATUS_READY
		self.sleepTime       = 0.2
		self.systemPool      = JobPool()
		self.userPool        = JobPool()
		self.closePool       = ClosePool()

		self._settings       = DCSettings()
		self._localAddress   = None
		self._searchListener = None
		self._searchWorker   = DCSearchWorker(self)
		self._xferListener   = None
		self._xferWorker     = DCXferWorker()
		self._queue          = DCQueue(self)
		self._localList      = None
		self._usedSlots      = 0
		self._infoMask       = None
		self._xferNicks      = {}

		self.systemPool.addJob(self._searchWorker)
		self.systemPool.addJob(self._xferWorker)
		self.systemPool.addJob(self._queue)

		atexit.register(self.stop)

	def start(self):
		if self._settings == None:
			raise SettingsException
		
		start_new_thread(self.run, ())

	def stop(self):
		if self.status == self.STATUS_READY:
			return
		elif self.status > self.STATUS_READY:
			self.status = self.STATUS_STOPPING

		if self.status < self.STATUS_READY:
			while self.status != self.STATUS_STOPPED:
				sleep(self.sleepTime)
		
			try:
				if self._queue:
					self._queue.save()
			except IOError: pass
			
			DCLocalListWriter.cleanup()
	
	def isAlive(self):
		return self.status == self.STATUS_RUNNING or self.status == self.STATUS_STOPPING

	def run(self):
		self.status = self.STATUS_RUNNING
		
		# Favorite autoconnection
		for addr, nick, password in [ (f[0], f[2], f[3]) for f in self._settings.getFavorites() if f[1] ]:
			self.addJob(DCHub(addr, nick, password))

		# Main loop
		while self.status == self.STATUS_RUNNING:
			self.expireXferNicks()
			self.systemPool.run()
			self.userPool.run()
			self.closePool.run()
			sleep(self.sleepTime)

		# Stop everything
		for job in self.systemPool.getJobs():
			self.closePool.addJob(job)

		for job in self.userPool.getJobs():
			self.closePool.addJob(job)

		while self.closePool.hasJobs():
			self.closePool.run()
			sleep(self.sleepTime)

		self.status = self.STATUS_STOPPED

	def addJob(self, job):
		self.userPool.addJob(job)

		if isinstance(job, DCHub):
			for listener in self.getListeners():
				listener.onNewHub(job)

	def removeJob(self, job):
		self.userPool.removeJob(job)

	def loadQueue(self):
		try: self._queue.load()
		except IOError: pass

	def loadSettings(self, path):
		self._settings.load(path)
		self.refreshSettings()
	
	def saveSettings(self, path):
		if self._settings.isValid():
			self._settings.save(path)
	
	def refreshSettings(self):
		# IP address
		if self._settings.ip != None:
			self._localAddress = self._settings.ip
		else:
			self._localAddress = getLocalIP()
		
		self.refreshLocalList(True)
		
		# Set bandwidth limits
		self._xferWorker.setUploadBandwidth(self._settings.getUploadBandwidth())
		self._xferWorker.setDownloadBandwidth(self._settings.getDownloadBandwidth())
		
		# Destroy listeners
		if self._searchListener != None:
			self.systemPool.removeJob(self._searchListener)
			self.closePool.addJob(self._searchListener)
			self._searchListener = None

			self.systemPool.removeJob(self._xferListener)
			self.closePool.addJob(self._xferListener)
			self._xferListener = None
		
		# If we're in active mode, create listeners
		if self._settings.active == True:
			port = self._settings.port
			self._searchListener = DCSearchListener(port, self._searchWorker)
			self.systemPool.addJob(self._searchListener)
			
			self._xferListener = DCXferListener(port)
			self.systemPool.addJob(self._xferListener)
		
		self.updateInfo(False)
	
	def refreshLocalList(self, cache = False):
		self.opLock.acquire()
		
		try:
			if cache and self._localList:
				paths = self._localList.getPaths()
				for path in self._settings.sharedDirs:
					paths.remove(path)
				
				if len(paths) == 0:
					return self.opLock.release()
		except ValueError:
			pass
		
		# Build local lists and compute share size
		self._localList = DCLocalList()
		for path in self._settings.sharedDirs:
			self._localList.scan(path)
		self._localList.freeze()

		self._settings.user.share = self._localList.getSize()
		DCLocalListWriter.write(self._localList)
		
		self.opLock.release()
	
	def refreshIP(self, ip, source):
		if self._settings.ip: return
		
		if ip != self._localAddress:
			self._localAddress = ip
			for h in [ h for h in self.getHubs() if h != source ]:
				h.reconnect(True)
	
	def updateInfo(self, silent = True):
		self.opLock.acquire()

		user = self._settings.user
		info = '$ALL %s '    # placeholder for user nick
		if user.description != None:
			info += user.description

		#DC++-like tag
		if self._settings.active:
			modeFlag = 'A'
		else:
			modeFlag = 'P'

		hubs = [ j for j in self.userPool.getJobs() if isinstance(j, DCHub) and j.isLogged() ]
		userNum, regNum, opNum = 0, 0, 0
		for hub in hubs:
			if hub.isOp():
				opNum += 1
			elif hub.isRegistered():
				regNum += 1
			else:
				userNum += 1
		
		info += '<%s %s><++ V:%s,M:%s,H:%d/%d/%d,S:%d' % (self.PYDCLIB_ID, self.PYDCLIB_VERSION, self.PYDCLIB_DCPP_VERSION, modeFlag, userNum, regNum, opNum, self._settings.getSlotNum())
		
		upBand = self._settings.getUploadBandwidth()
		if upBand > 0:
			info += ',L:%d>' % round(upBand/1024)
		else:
			info += '>' 

		# Other user info
		info += '$ $' + DCUser.CONN_NAMES[user.connection]
		info += '%s$'         # placeholder for user status flag

		if user.email != None:
			info += user.email
		info += "$"

		if user.share == -1:
			info += "0"
		else:
			info += str(user.share)
		info += "$"

		self._infoMask = info
		self.opLock.release()

		for hub in hubs:
			hub.updateInfo(silent)
	
	def deregisterAllListeners(self):
		for job in self.systemPool.getJobs():
			if isinstance(job, EventGenerator):
				job.deregisterAllListeners()

	def getSettings(self):
		if self._settings == None: raise SettingsException()
		return self._settings

	def getLocalAddress(self):
		if self._settings == None: raise SettingsException()
		return self._localAddress

	def getSearchWorker(self):
		return self._searchWorker

	def getXferWorker(self):
		return self._xferWorker

	def getQueue(self):
		return self._queue

	def getLocalList(self):
		self.opLock.acquire()
		l = self._localList
		self.opLock.release()
		return l

	def getUsedSlots(self):
		self.opLock.acquire()
		slots = self._usedSlots
		self.opLock.release()
		return slots

	def getInfoMask(self):
		return self._infoMask

	def getJobs(self):
		return self.userPool.getJobs()

	def getXfers(self):
		return self._xferWorker.getXfers()

	def getHubs(self):
		return [ j for j in self.userPool.getJobs() if isinstance(j, DCHub) ]

	def getHubByName(self, name):
		l = [ h for h in self.getHubs() if h.getName() == name ]
		if len(l):
			return l[0]
		else:
			return None

	def getUsersByNick(self, nick):
		aux = []
		for hub in self.getHubs():
			user = hub.getUserByNick(nick)
			if user != None: aux.append(user)
		return aux

	###################
	# Private members #
	###################
	
	def registerXferNick(self, local, remote):
		self._xferNicks[remote] = (local, time())
	
	def getXferNick(self, remote):
		nick = self._xferNicks[remote][0]
		del self._xferNicks[remote]
		return nick
	
	def expireXferNicks(self):
		t = time()
		aux = {}
		for k, v in self._xferNicks.iteritems():
			if t - v[1] < self.XFER_GRACE:
				aux[k] = v
		self._xferNicks = aux
	
	def acquireSlot(self):
		try:
			self.opLock.acquire()
			if self._usedSlots < self._settings.getSlotNum():
				self._usedSlots += 1
				return True
			else:
				return False
		finally:
			self.opLock.release()

	def releaseSlot(self):
		self.opLock.acquire()
		if self._usedSlots > 0:
			self._usedSlots -= 1
		self.opLock.release()
	
	settings = property(getSettings)
	localAddress = property(getLocalAddress)
	searchWorker = property(getSearchWorker)
	xferWorker = property(getXferWorker)
	queue = property(getQueue)
	jobs = property(getJobs)
	hubs = property(getHubs)
	xfers = property(getXfers)
	usedSlots = property(getUsedSlots)

class SettingsException(Exception): pass

#Global object
sys.modules['__main__'].dcWorker = DCWorker()

def getWorker():
	return sys.modules['__main__'].dcWorker
