# This file is part of pyDC software
# Copyright 2002-2004 Anakim Border <aborder@users.sourceforge.net>
#
# pyDC is released under the terms of GPL licence.

from array import array
from random import randint
from sets import ImmutableSet
from time import time
from weakref import proxy

from DCChannel import *
from DCChallenge import solveChallenge
from DCDownload import *
from DCUpload import *
import DCWorker

class DCXfer(DCChannel):
	TYPE_UNKNOWN = 0
	TYPE_DOWNLOAD = 1
	TYPE_UPLOAD = 2
	
	DOWNLOAD_ROLLBACK = 4096
	
	def __init__(self, inbound, param, nick = None):
		DCChannel.__init__(self)
		
		self._type = self.TYPE_UNKNOWN
		self._inbound = inbound
		self._supports = ImmutableSet()
		self._localNick = nick
		self._remoteNick = None
		self._item = None
		self._sizeRemaining = -1
		self._progress = 0
		self._compressed = False

		self.file = None
		self.notified = False
		self._speed = SpeedManager(self)
		
		if inbound:
			self.attach(param)
		else:
			self.connect(param)
	
	def onConnection(self):
		self.setHandler(DCXferHandshakeHandler())
	
	def onDisconnection(self):
		self.xferClosed()
		
	def onSocketError(self, error):
		self.xferClosed()
	
	def setLocalNick(self, nick):
		assert self._localNick == None
		self._localNick = nick
	
	def setSupports(self, exts):
		self._supports = ImmutableSet(exts)
	
	def handshakeCompleted(self):
		if self._type == self.TYPE_DOWNLOAD:
			self.setHandler(DCDownloadSetupHandler())
		else:
			self.setHandler(DCUploadSetupHandler())
	
	def notify(self):
		if not self.notified:
			DCWorker.getWorker().getXferWorker().onNewXfer(self)
			self.notified = True
	
	def xferClosed(self):
		if self.notified:
			DCWorker.getWorker().getXferWorker().onXferClosed(self)
			self.notified = False
	
	def nextItem(self):
		self._item = None
		self._sizeRemaining = -1
		self._progress = 0
		self.clearCleanup()
		
		if self.file:
			self.file.close()
			self.file = None
		
		self.xferClosed()
		self.handshakeCompleted()

	# Properties
	def getType(self):
		return self._type
		
	def isInbound(self):
		return self._inbound
	
	def supports(self, ext):
		return ext in self._supports
	
	def supportsSegments(self):
		return 'GetZBlock' in self._supports or 'GetTestZBlock' in self._supports
	
	def getLocalNick(self):
		return self._localNick

	def getRemoteNick(self):
		return self._remoteNick

	def getPath(self):
		return self._item.getPath()
	
	def getItem(self):
		return self._item

	def getSizeRemaining(self):
		return self._sizeRemaining

	def getProgress(self):
		return self._progress

	def getSpeed(self):
		return self._speed.getSpeed()
		
	def isCompressed(self):
		return self._compressed
	
	type = property(getType)
	inbound = property(isInbound)
	localNick = property(getLocalNick)
	remoteNick = property(getRemoteNick)
	item = property(getItem)
	sizeRemaining = property(getSizeRemaining)
	progress = property(getProgress)
	speed = property(getSpeed)
	compressed = property(isCompressed)

class DCXferHandshakeHandler(DCChannelHandler):
	HANDSHAKE_TIMEOUT = 60
	LOCK_TEMPLATE = 'EXTENDEDPROTOCOLABCABCABCABCABCABC Pk=%s'
	LOCK_PK       = 'ABCABCABCABCABCABCABC'

	def __init__(self):
		cmds = { 'MyNick' :    ( 1, DCXferHandshakeHandler.onNick),
		         'Lock' :      ( 2, DCXferHandshakeHandler.onLock),
		         'Supports' :  (-1, DCXferHandshakeHandler.onSupports),
		         'Direction' : ( 2, DCXferHandshakeHandler.onDirection),
				 'Key' :       (-1, DCXferHandshakeHandler.onKey) }
		DCChannelHandler.__init__(self, cmds)
		
		self.timestamp = None
		self.rval = randint(0, 32766)
		self.preferredType = None
	
	def sendInfo(self):
		self.channel.sendCmd('MyNick', self.channel.getLocalNick())
		
		#Generate Pk using pyDClib id and version
		w = DCWorker.getWorker()
		pk = w.PYDCLIB_ID.upper() + w.PYDCLIB_VERSION.upper()
		l, m = len(pk), len(self.LOCK_PK)
		if l > m:
			pk = pk[:m]
		elif l < m: 
			pk = pk + self.LOCK_PK[l:]
		
		self.channel.sendCmd('Lock', self.LOCK_TEMPLATE % pk)
	
	def poll(self):
		if time() - self.timestamp > self.HANDSHAKE_TIMEOUT:
			raise ValueError, '[xfer] handshake timeout'
		
		DCChannelHandler.poll(self)
	
	def onAttach(self, xfer):
		DCChannelHandler.onAttach(self, xfer)
		self.timestamp = time()
		if not self.channel.isInbound(): self.sendInfo()
	
	def onNick(self, nick):
		self.channel._remoteNick = nick
		if self.channel.isInbound():
			self.channel.setLocalNick(DCWorker.getWorker().getXferNick(nick))
			self.sendInfo()
	
	def onLock(self, key, pk):
		if self.channel.getRemoteNick() == None:
			raise ValueError
		
		if key[:16] == 'EXTENDEDPROTOCOL':
			self.channel.sendCmd('Supports', 'BZList', 'GetTestZBlock', 'GetZBlock')
		
		item, source = DCWorker.getWorker().getQueue().getItemByNick(self.channel.getRemoteNick())
		if item == None:
			self.preferredType = DCXfer.TYPE_UPLOAD
			self.channel.sendCmd('Direction', 'Upload', str(self.rval))
		else:
			self.preferredType = DCXfer.TYPE_DOWNLOAD
			self.channel.sendCmd('Direction', 'Download', str(self.rval))

		self.channel.sendCmd('Key', solveChallenge(key))
		
	def onSupports(self, data):
		if data:
			self.channel.setSupports(data.split(' '))
		
	def onDirection(self, direction, rval):
		if direction == 'Upload':
			type = DCXfer.TYPE_DOWNLOAD
		elif direction == 'Download':
			type = DCXfer.TYPE_UPLOAD
		else:
			raise InvalidCommand

		if type == self.preferredType:
			self.channel._type = type
		else:
			if self.rval > int(rval):
				self.channel._type = self.preferredType
			else:
				self.channel._type = type

	def onKey(self, data):
		if self.channel._type == self.channel.TYPE_UNKNOWN:
			self.channel.sendCmd('Error')
			raise ValueError

		self.channel.handshakeCompleted()

class SpeedManager:
	SPEED_RECORDS = 30
	MINSPEED_TIMEOUT = 30

	def __init__(self, xfer):
		self.xfer = proxy(xfer)
		self.worker = DCWorker.getWorker().getXferWorker()
		self.records = [0]
		self.minSpeed = 1.0

	def init(self, time):
		self.records = array('f', [0] * self.SPEED_RECORDS)
		self.recordPos = 0
		self.timestamp = time
		self.minSpeedCount = 0
		self.lastSpeed = 10240

	def getBandwidth(self, time):
		bw = self.worker.xferBandwidth(self.xfer._type)
		delta = (time - self.timestamp)
		
		if bw == 0:
			return delta * max(self.lastSpeed, 10240) * 10
		else:
			return delta * bw

	def usedBandwidth(self, time, netBytes, dataBytes):
		delta = time - self.timestamp
		netSpeed = netBytes / delta
		dataSpeed = dataBytes / delta
		self.worker.usedBandwidth(self.xfer._type, netSpeed)

		self.records[self.recordPos] = dataSpeed
		self.recordPos = (self.recordPos + 1) % len(self.records)
		self.timestamp = time

		if dataSpeed < self.minSpeed:
			self.minSpeedCount += delta
			if self.minSpeedCount >= self.MINSPEED_TIMEOUT:
				self.xfer.stop()
		else:
			self.minSpeedCount = 0

	def getSpeed(self):
		self.lastSpeed = sum(self.records) / len(self.records)
		return self.lastSpeed

	def getMinSpeed(self):
		return self.minSpeed

	def getLastTimestamp(self):
		return self.timestamp

	def setMinSpeed(self, speed):
		self.minSpeed = speed
		self.minSpeedCount = 0
