# This file is part of pyDC software
# Copyright 2002-2004 Anakim Border <aborder@users.sourceforge.net>
#
# pyDC is released under the terms of GPL licence.

from copy import copy
from EventGenerator import *
from Job import Job
from threading import Lock

from DCXfer import DCXfer

class DCXferWorker(object, Job, EventGenerator):
	def __init__(self):
		EventGenerator.__init__(self)

		self._xfers = []
		self.limits = { DCXfer.TYPE_DOWNLOAD : SpeedLimit(self),
		                DCXfer.TYPE_UPLOAD   : SpeedLimit(self) }
		self.opLock = Lock()

	def poll(self):
		for limit in self.limits.itervalues():
			limit.reset()

	def stop(self):
		self.poll = self.stoppedPoll

	def isAlive(self):
		return self.poll != self.stoppedPoll

	def stopXfer(self, item, source = None):
		self.opLock.acquire()
		for xfer in self._xfers:
			if xfer.getItem() == item:
				if source == None or xfer.getRemoteNick() == source.nick:
					xfer.stop()
		self.opLock.release()

	def xferBandwidth(self, type):
		self.opLock.acquire()
		bps = self.limits[type].xferBandwidth()
		self.opLock.release()
		return bps

	def usedBandwidth(self, type, bytes):
		self.opLock.acquire()
		self.limits[type].usedBandwidth(bytes)
		self.opLock.release()

	def onNewXfer(self, xfer):
		self.opLock.acquire()
		self._xfers.append(xfer)
		self.limits[xfer.getType()].incXferNum()
		self.opLock.release()

		self.lock.acquire()
		for listener in self.getListeners():
			try: listener.onNewXfer(xfer)
			except: pass
		self.lock.release()

	def onXferUpdate(self, xfer):
		self.lock.acquire()
		for listener in self.getListeners():
			try: listener.onXferUpdate(xfer)
			except: pass
		self.lock.release()

	def onXferClosed(self, xfer):
		self.opLock.acquire()
		self._xfers.remove(xfer)
		self.limits[xfer.getType()].decXferNum()
		self.opLock.release()

		self.lock.acquire()
		for listener in self.getListeners():
			try: listener.onXferClosed(xfer)
			except: pass
		self.lock.release()

	def getXfers(self):
		self.opLock.acquire()
		xfers = copy(self._xfers)
		self.opLock.release()
		return xfers

	def getDownloadBandwidth(self):
		self.opLock.acquire()
		bps = self.limits[DCXfer.TYPE_DOWNLOAD].getBandwidth()
		self.opLock.release()
		return bps

	def setDownloadBandwidth(self, bps):
		self.opLock.acquire()
		self.limits[DCXfer.TYPE_DOWNLOAD].setBandwidth(bps)
		self.opLock.release()

	def getUploadBandwidth(self):
		self.opLock.acquire()
		bps = self.limits[DCXfer.TYPE_UPLOAD].getBandwidth()
		self.opLock.release()
		return bps

	def setUploadBandwidth(self, bps):
		self.opLock.acquire()
		self.limits[DCXfer.TYPE_UPLOAD].setBandwidth(bps)
		self.opLock.release()

	#=================#
	# Private section #
	#=================#

	def stoppedPoll(self):
		return True
	
	xfers = property(getXfers)
	downloadBandwidth = property(getDownloadBandwidth, setDownloadBandwidth)
	uploadBandwidth = property(getUploadBandwidth, setUploadBandwidth)

class SpeedLimit:
	def __init__(self, worker):
		self.worker = worker
		
		self.totalBandwidth = 0
		self.xferCount = 0
		self.xferBand = 0
		self.toLimit = 0
		self.unused = 0

	def incXferNum(self):
		self.xferCount += 1
		self.toLimit += 1
		self.calcBandwidth()

	def decXferNum(self):
		if self.xferCount > 0:
			self.xferCount -= 1
			self.calcBandwidth()

	def calcBandwidth(self):
		if self.xferCount > 0:
			self.xferBand = float(self.totalBandwidth) / self.xferCount

	def xferBandwidth(self):
		if self.totalBandwidth == 0: return 0
		return self.xferBand + (self.unused / self.toLimit)

	def usedBandwidth(self, bps):
		if self.totalBandwidth > 0:
			self.unused += self.xferBand - bps
		self.toLimit -= 1

	def reset(self):
		self.calcBandwidth()
		self.toLimit = self.xferCount
		self.unused = min(self.unused, self.totalBandwidth)

	def getBandwidth(self):
		return self.totalBandwidth

	def setBandwidth(self, bps):
		if bps < 0: raise ValueError
		self.totalBandwidth = bps
		self.calcBandwidth()
		self.unused = 0
