# This file is part of pyDC software
# Copyright 2002-2004 Anakim Border <aborder@users.sourceforge.net>
#
# pyDC is released under the terms of GPL licence.

from copy import copy
from time import time
from socket import *
from threading import Lock
from Job import *
import errno
import os
import re
import struct
import sys

class DNS:
	def __init__(self):
		self.servers = []
		self.serial = 0
		self.resolved = None
	
	def getNameservers(self):
		if len(self.servers) == 0:
			self._scan()
			if len(self.servers) == 0:
				raise DNSError, 'can\'t find nameservers'
		return copy(self.servers)
		
	def cycleServers(self):
		self.servers = self.servers[1:] + [ self.servers[0] ]
	
	def getSerial(self):
		serial = self.serial
		self.serial = (self.serial + 1) % 0xffff
		return serial
	
	def _scan(self):
		if os.name == 'posix':
			try:
				f = open('/etc/resolv.conf')
				p = re.compile('nameserver[ \t]*([^ \t\n]*)')
				self.servers = p.findall(f.read())
				f.close()
			except IOError:
				return
		elif os.name == 'nt':
			import win32dns
			self.servers = win32dns.getNameservers()
		
		self.serverIdx = 0

class DNSResolver(Job):
	def __init__(self, addr, timeout = 10):
		self.addr = addr
		self.timeout = timeout
		self.resolved = None
		
		self.servers = getDNS().getNameservers()
		self.serverIdx = 0
		self.serial = getDNS().getSerial()
		self.question = None
		self.questionLen = 0
		
		self.sock = None
		self.starttime = 0
		self.poll = self.pollSendQuestion
	
	def poll(self):
		pass
	
	def pollSendQuestion(self):
		if self.question == None:
			self.question, self.questionLen = self.buildQuestion()
		
		# Open socket & send
		try:
			self.sock = socket(AF_INET, SOCK_DGRAM)
			self.sock.setblocking(False)
			self.sock.connect((self.servers[self.serverIdx], 53))
			self.sock.send(self.question)
		except error:
			if not self.nextServer():
				raise DNSError, 'can\'t connect to nameserver'
			else:
				return False
		
		self.starttime = time()
		self.poll = self.pollWaitReply
	
	def pollWaitReply(self):
		try:
			self.reply = self.sock.recv(1024)
			if not self.parseReply():
				if not self.nextServer():
					raise DNSError, 'corrupted server reply'
				else:
					return False
			
			self.cleanup()
			if self.resolved == None:
				raise DNSError, 'address unknown'
		
		except error, e:
			if time() - self.starttime >= self.timeout:
				if not self.nextServer():
					raise DNSError, 'timeout while waiting reply'
				else:
					return False
			
			if self.isWouldBlock(e[0]):
				return False
			else:
				if not self.nextServer():
					raise DNSError, 'broken connection with server'
				else:
					return False
	
	def stop(self):
		self.cleanup()

	def isAlive(self):
		# cleanup() will change this function when
		# work is done.
		return True

	#####################
	#  Private members  #
	#####################
	
	def isWouldBlock(self, code):
		if os.name == 'posix':
			if code == errno.EAGAIN or code == errno.EALREADY or code == errno.EINPROGRESS:
				return True
		elif os.name == 'nt':
			if code == errno.WSAEWOULDBLOCK:
				return True
		
		return False
	
	def cleanup(self):
		if self.sock != None:
			self.sock.close()
		
		self.isAlive = lambda:False
		self.poll = lambda:True
	
	def nextServer(self):
		self.cleanup()
		getDNS().cycleServers()
		
		self.serverIdx += 1
		if self.serverIdx < len(self.servers):
			self.poll = self.pollSendQuestion
			return True
		else:
			self.poll = self.pollStopped
			return False

	def buildQuestion(self):
		qlen = 0
	
		# Header
		buf  = self.pack16bit(self.serial)
		buf += self.pack16bit(1<<8)
		buf += self.pack16bit(1)
		buf += self.pack16bit(0)
		buf += self.pack16bit(0)
		buf += self.pack16bit(0)
		
		# Question
		for label in self.addr.split('.'):
			llen = len(label)
			if llen > 63: raise DNSError, 'label too long'
			buf += chr(llen) + label
			qlen += llen + 1
		
		buf += '\0'
		buf += self.pack16bit(1)
		buf += self.pack16bit(1)
		qlen += 5
		
		return buf, qlen
	
	def parseReply(self):
		self.replyPos = 0

		#Header
		if self.get16bit() != self.serial:
			raise DNSError, "unexpected ID"
		flags = self.get16bit()
		if (flags>>15)&1 != 1: raise DNSError, "reply error"
		if (flags>>0)&0xF != 0:
			if (flags>>0)&0xF == 3: return 1
			else: raise DNSError, "server failure"
		#print self.get16bit(), self.get16bit(), self.get16bit(), self.get16bit()
		question = self.get16bit()
		answer = self.get16bit()
		authority = self.get16bit()
		self.replyPos += 2

		if question == 1: self.replyPos += self.questionLen
		if answer == 0 and authority == 0: raise DNSError, "reply error"

		rrs = []
		for i in range(0, answer+authority):
			type, data = self.parseRR()
			if type == 1:
				self.resolved = data
				return 1
			else:
				rrs.append((type, data))

		addrs = filter(lambda i: i[0] == 2, rrs)
		if len(addrs) != 0:
			self.server = addrs[0][1]
			return 0

		return 1

	def parseRR(self):
		name = self.getname()
		type = self.get16bit()
		if self.get16bit() != 1: raise DNSError, "reply error"
		self.replyPos += 4
		rlen = self.get16bit()

		if type == 1:
			data = inet_ntoa(self.reply[self.replyPos:self.replyPos+4])
			self.replyPos += rlen
		elif type == 2:
			data = self.getname()
		elif type == 5:
			data = self.getname()
		else:
			raise DNSError, "reply error"

		return type, data

	def getname(self):
		labels = []

		while 1:
			llen = ord(self.get8bit())

			if llen == 0: break
			elif llen & 0xC0 == 0xC0:
				j = ord(self.get8bit())
				pointer = ((llen<<8) | j) & ~0xC000

				realPos = self.replyPos
				self.replyPos = pointer
				labels.append(self.getname())
				self.replyPos = realPos

				break
			else:
				labels.append(self.reply[self.replyPos:self.replyPos+llen])
				self.replyPos += llen

		if len(labels) == 0: raise DNSError, "RR name is null"
		name = ""
		for i in range(0, len(labels)-1):
			name += labels[i] + '.'
		name += labels[-1]
		return name
	
	def get8bit(self):
		bits = self.reply[self.replyPos]
		self.replyPos += 1
		return bits
	
	def get16bit(self):
		bits = self.reply[self.replyPos:self.replyPos+2]
		self.replyPos += 2
		return self.unpack16bit(bits)
	
	def pack16bit(self, n):
		return struct.pack('!H', n)
	
	def unpack16bit(self, s):
		return struct.unpack('!H', s)[0]

class DNSError(Exception): pass

#Global object
sys.modules['__main__'].dns = DNS()

def getDNS():
	return sys.modules['__main__'].dns

if __name__ == "__main__":
	r = DNSResolver(sys.argv[1])
	while not r.poll(): pass
	print "Resolved:", r.resolved
