/*This file is part of pyDC software
Copyright 2002 Anakim Border <aborder@users.sourceforge.net>

pyDC is released under the terms of GPL licence.*/

#include "codec.h"

BitStream::BitStream(unsigned char* data, unsigned int size) : m_data(data), m_pos(0), m_num(0)
{
	if (size == static_cast<unsigned int>(-1))
		m_end = reinterpret_cast<unsigned char*>(-1);
	else
		m_end = &data[size];
}

unsigned int BitStream::getBit()
{
	if (m_data >= m_end)
		throw IndexError();

	char bit = *m_data >> (m_pos & 7) & 1;
	if (m_pos == 7)
	{
		m_pos = 0;
		m_data++;
		m_num++;
	}
	else
		m_pos++;

	return bit;
}

unsigned int BitStream::getBits(unsigned int num)
{
	unsigned int value = 0;
	for (unsigned int i = 0; i < num; i++)
		value = (value << 1) | getBit();
		//value |= getBit() << i;

	return value;
}

void BitStream::setBit(const unsigned char value)
{
	if (m_data >= m_end)
		throw IndexError();

	if (m_pos == 0) *m_data = 0;
	*m_data |= (value & 1) << m_pos;

	m_pos++;
	if (m_pos > 7)
	{
		m_pos = 0;
		m_data++;
		m_num++;
	}

}

void BitStream::setBits(const unsigned int value, const int len)
{
	for (int i = len-1; i >= 0; i--)
		setBit(value >> i);
}

void BitStream::pad()
{
	if (m_pos)
	{
		m_pos = 0;
		m_data++;
		m_num++;
	}
}

unsigned int BitStream::getByteNum() const
{
	return m_num + (m_pos != 0 ? 1 : 0);
}

void Encoder::encode(const char* data)
{
	PatternInfo dict[256];
	unsigned int i;
	for (i = 0; i < 256; i++)
		dict[i].value = static_cast<unsigned char>(i);

	const unsigned int len = strlen(data);
	for (i = 0; i < len; i++)
		dict[static_cast<unsigned char>(data[i])].count++;

	//Build tree leaves
	list<TreeNode*> nodes;
	for (i = 0; i < 256; i++)
	{
		if (dict[i].count > 0)
		{
			TreeNode *n = new TreeNode(&dict[i]);
			insertNode(nodes, *n);
		}
	}

	//Build Huffman tree
	while (nodes.size() > 1)
	{
		//print_tree(nodes);

		TreeNode *l = nodes.front();
		nodes.pop_front();
		TreeNode *r = nodes.front();
		nodes.pop_front();

		TreeNode *n = new TreeNode(l, r);
		insertNode(nodes, *n);
	}

	//Build patterns
	nodes.front()->setPattern();
	TreeNode *root = nodes.front();
	nodes.pop_front();
	delete root;

	//Write header
	unsigned char *buf = new unsigned char[11+512+1024+len];

	buf[0] = 'H';
	buf[1] = 'E';
	buf[2] = '3';
	buf[3] = 0xD;

	//Parity
	for(i = 0, buf[4] = 0; i < len; i++)
		buf[4] ^= static_cast<unsigned char>(data[i]);

	buf[5] = len & 255;
	buf[6] = (len >> 8) & 255;
	buf[7] = (len >> 16) & 255;
	buf[8] = (len >> 24) & 255;

	//Pattern info
	unsigned char *p = &buf[11];
	unsigned int patternNum = 0;
	for (i = 0; i < 256; i++)
	{
		if (dict[i].count)
		{
			p[patternNum*2] = dict[i].value;
			p[patternNum*2+1] = dict[i].len;
			patternNum++;
		}
	}

	buf[9]  = patternNum & 255;
	buf[10] = (patternNum >> 8) & 255;
	p += patternNum * 2;

	//Pattern list
	BitStream b(p, 1024+len);
	for (i = 0; i < 256; i++)
	{
		if (dict[i].count)
			b.setBits(dict[i].pattern, dict[i].len);
	}
	b.pad();

	//Encode data
	for (i = 0; i < len; i++)
	{
		unsigned char idx = data[i];
		b.setBits(dict[idx].pattern, dict[idx].len);
	}

	//Trim buffer
	m_len = 11 + patternNum*2 + b.getByteNum();
	m_encoded = new char[m_len];
	memcpy(m_encoded, buf, m_len);
	delete buf;
}

void Encoder::insertNode(list<TreeNode*> &l, TreeNode &node)
{
	list<TreeNode*>::iterator i = l.begin(), e = l.end();
	for (; i != e; i++)
	{
		if (node.weight < (*i)->weight)
			break;
		else if (node.weight == (*i)->weight && node.l == NULL)
			break;
	}

	l.insert(i, &node);
}

/*void Encoder::print_tree(list<TreeNode*> &nodes) const
{
	list<TreeNode*>::iterator i = nodes.begin(), e = nodes.end();
	for (; i != e; i++)
	{
		printf("occur: %ld, left=%p, right=%p, val=%c\n",
			(*i)->weight, (*i)->l, (*i)->r, (*i)->v != NULL ? (*i)->v->value : 0);
	}
}*/

void Encoder::TreeNode::setPattern(unsigned int base, unsigned char len)
{
	if (l)
	{
		len += 1;
		l->setPattern(base << 1, len);
		r->setPattern((base << 1) + 1, len);
	}
	else
	{
		v->pattern = base;
		v->len = len;
	}
}

bool Decoder::decode(char* data, unsigned int size)
{
	if (size < 10) return false;
	if (data[0] != 'H' || data[1] != 'E' || data[2] != '3' || data[3] != 0xD)
		return false;

	unsigned int bytesToProduce;
	bytesToProduce  = static_cast<unsigned char>(data[8]);
	bytesToProduce <<= 8;
	bytesToProduce |= static_cast<unsigned char>(data[7]);
	bytesToProduce <<= 8;
	bytesToProduce |= static_cast<unsigned char>(data[6]);
	bytesToProduce <<= 8;
	bytesToProduce |= static_cast<unsigned char>(data[5]);

	unsigned int patternNum;
	patternNum  = static_cast<unsigned char>(data[10]);
	patternNum <<= 8;
	patternNum |= static_cast<unsigned char>(data[9]);

	PatternInfo* infos = new PatternInfo[patternNum];
	unsigned int i = 0, pos = 11;
	for (; i < patternNum; i++)
	{
		infos[i].value = static_cast<unsigned char>(data[pos++]);
		infos[i].length = static_cast<unsigned char>(data[pos++]);
	}

	unsigned int bytes = 0, listLen = 0;
	for (i = 0; i < patternNum; i++)
	{
		bytes += infos[i].length;
		if (infos[i].length > listLen)
			listLen = infos[i].length;
	}
	listLen = 1 << (listLen + 1);

	unsigned char *patternList = new unsigned char[listLen];
	memset(patternList, 0, listLen);

	BitStream bs(reinterpret_cast<unsigned char*>(&data[pos]), listLen);
	for (i = 0; i < patternNum; i++)
	{
		unsigned int p = bs.getBits(infos[i].length);
		patternList[(1 << infos[i].length) + p] = infos[i].value;
	}
	delete infos;

	pos += bytes / 8 + (bytes % 8 ? 1 : 0);
	BitStream ds(reinterpret_cast<unsigned char*>(&data[pos]));
	delete m_decoded;
	m_decoded = new char[bytesToProduce+1];
	for (i = 0; i < bytesToProduce; i++)
	{
		pos = ds.getBit();
		bytes = 1;

		while (true)
		{
			unsigned int p = (1 << bytes) + pos;
			if (p >= listLen) return false;
			if (patternList[p] != 0)
				break;

			pos = (pos << 1) | ds.getBit();
			bytes++;
		}

		m_decoded[i] = patternList[(1 << bytes) + pos];
	}
	m_decoded[bytesToProduce] = 0;
	delete patternList;

	return true;
}
